/*
 *   (C) Copyright IBM Corp. 2006
 *   (C) Copyright ALT Linux Ltd. 2011
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Module: BtrFS FSIM
 * File: evms2/engine/plugins/btrfs/fsimbtrfs.c
 */

#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>
#include <libgen.h>

#include <plugin.h>

#include <btrfsutil.h>

#include "fsimbtrfs.h"

plugin_record_t		btrfs_plugrec;
plugin_record_t		*my_plugin_record=&btrfs_plugrec;
engine_functions_t	*EngFncs;
boolean			have_mkfs_prog = FALSE;

static int btrfs_do_destroy_subvols(logical_volume_t * vol);

void subvol_name_to_btrfs_name(char * path, char * volume_name, char * name)
{
	LOG_ENTRY();

	// volume_name starts with "/dev/evms"
	snprintf(name, EVMS_NAME_SIZE, "%s/%s/%s", BTRFS_DEV_DIRECTORY, volume_name + 10, path);

	LOG_EXIT_VOID();
}

int get_btrfs_super_block(logical_volume_t *volume, struct btrfs_super_block *sb_ptr)
{
	int fd, rc;

	fd = EngFncs->open_volume(volume, O_RDONLY, 0);
	if (fd<0) return EBUSY;

	if (EngFncs->read_volume(volume, fd, sb_ptr, BTRFS_SIZE_OF_SUPER_BLOCK, BTRFS_SUPER_LOC)==BTRFS_SIZE_OF_SUPER_BLOCK) rc=0;
	else rc=ENOENT;

	EngFncs->close_volume(volume, fd);
	return rc;
};

int try_run(char *prog_name)
{
	int rc;
	char *argv[3];
	pid_t pidm;
	int status;
	int fds[2];

	rc = pipe(fds);
	if (rc) return errno;

	argv[0] = prog_name;
	argv[1] = NULL;

	pidm = EngFncs->fork_and_execvp(NULL, argv, NULL, NULL, NULL);
	if (pidm != -1)
	{
		waitpid(pidm, &status, 0);
		if (!WIFEXITED(status) || (WEXITSTATUS(status) == ENOENT)) rc = EINTR;
	}
	else rc = errno;

	close(fds[0]);
	close(fds[1]);

	return rc;
};

/* sort before creating */
static int sort_subvols_asc(void * thing1, void * thing2, void * UNUSED(user_data)) {
	subvolume_t * subvol1 = (subvolume_t *) thing1;
	subvolume_t * subvol2 = (subvolume_t *) thing2;

	if (strstr(subvol1->name, subvol2->name)) return 1;
	if (strstr(subvol2->name, subvol1->name)) return -1;
	return 0;
}

/* sort before removing */
static int sort_subvols_desc(void * thing1, void * thing2, void * UNUSED(user_data)) {
	subvolume_t * subvol1 = (subvolume_t *) thing1;
	subvolume_t * subvol2 = (subvolume_t *) thing2;

	if (strstr(subvol1->name, subvol2->name)) return -1;
	if (strstr(subvol2->name, subvol1->name)) return 1;
	return 0;
}

static int btrfs_setup(engine_functions_t *engine_function_table)
{
	EngFncs = engine_function_table;
	if (try_run(BTRFS_MKFS_PROG) == 0) have_mkfs_prog = TRUE;
	return 0;
}

static void btrfs_cleanup(void)
{
	int rc = 0;
	list_anchor_t global_volumes;
	list_element_t vol_list_iter;
	logical_volume_t * volume;
	list_element_t subvol_list_iter;

	LOG_ENTRY();

	rc = EngFncs->get_volume_list(my_plugin_record, NULL, 0, &global_volumes);
	if (!rc)
	{
		LIST_FOR_EACH(global_volumes, vol_list_iter, volume)
		{
			if (volume->private_data)
			{
				EngFncs->engine_free(volume->private_data);
				volume->private_data = NULL;
			}

			if (volume->subvolumes)
			{
				void * thing;
				LIST_FOR_EACH(volume->subvolumes, subvol_list_iter, thing)
				{
					EngFncs->free_subvolume(thing);
				}

				EngFncs->destroy_list(volume->subvolumes);
				volume->subvolumes = NULL;
			}
		}
		EngFncs->destroy_list(global_volumes);
	}
	LOG_EXIT_VOID();

	have_mkfs_prog = FALSE;
};

static int btrfs_mount(logical_volume_t *volume, char *mountpoint)
{
	int rc=0;
	int status;
	char *cmd[6];
	cmd[0] = "mount";
	cmd[1] = "-t";
	cmd[2] = "btrfs";
	cmd[3] = EVMS_GET_DEVNAME(volume);
	cmd[4] = mountpoint;
	cmd[5] = NULL;

	LOG_ENTRY();
	int fds[2];
	fds[0] = 0;
	fds[1] = open("/dev/null", O_WRONLY);

	LOG_DEBUG("Mounting volume %s at %s\n", EVMS_GET_DEVNAME(volume), mountpoint);
	pid_t pidm;
	pidm = EngFncs->fork_and_execvp(volume, cmd, NULL, fds, fds);
	if (pidm != -1)
	{
		waitpid(pidm, &status, 0);
		if (WIFEXITED(status)) rc=WEXITSTATUS(status);
		else rc=EINTR;
		volume->mount_point = mountpoint;
	}
	else rc=errno;

	close(fds[1]);

	LOG_EXIT_INT(rc);
	return rc;
}

static int btrfs_umount(logical_volume_t *volume)
{
	int rc=0;
	int status;
	char *cmd[3];
	cmd[0] = "umount";
	cmd[1] = EVMS_GET_DEVNAME(volume);
	cmd[2] = NULL;

	LOG_ENTRY();
	int fds[2];
	fds[0] = 0;
	fds[1] = open("/dev/null", O_WRONLY);

	LOG_DEBUG("Unmounting volume %s\n", EVMS_GET_DEVNAME(volume));
	pid_t pidm;
	pidm = EngFncs->fork_and_execvp(volume, cmd, NULL, fds, fds);
	if (pidm != -1)
	{
		waitpid(pidm, &status, 0);
		if (WIFEXITED(status)) rc=WEXITSTATUS(status);
		else rc=EINTR;
		volume->mount_point = NULL;
	}
	else rc=errno;

	close(fds[1]);

	LOG_EXIT_INT(rc);
	return rc;
}

static int btrfs_get_subvols(logical_volume_t *volume)
{
	int rc = 0;
	char *subvol_path;
	subvolume_t *subvol;
	enum btrfs_util_error btrfs_rc;
	char *path = "/mnt/destination";
	struct btrfs_util_subvolume_iterator *iter;
	char full_path[EVMS_NAME_SIZE+1];

	LOG_ENTRY();
	rc = btrfs_mount(volume, path);

	if (rc != 0)
	{
		goto error;
	}

	volume->subvolumes = EngFncs->allocate_list();

	/* do fill allocated list */
	btrfs_rc = btrfs_util_create_subvolume_iterator(path, 5 /*BTRFS_FS_TREE_OBJECTID*/, 0, &iter);

	if (btrfs_rc != BTRFS_UTIL_OK)
	{
		LOG_ERROR("%s\n", btrfs_util_strerror(btrfs_rc));
		goto error;
	}

	struct btrfs_util_subvolume_info info;

	while (!(btrfs_rc = btrfs_util_subvolume_iterator_next_info(iter, &subvol_path, &info))) {
		subvol_name_to_btrfs_name(subvol_path, volume->name, full_path);

		if (EngFncs->allocate_subvolume(full_path, &subvol)) {
			rc = ENOMEM;
			break;
		}
		subvol->volume = volume;
		strncpy(subvol->short_name, subvol_path, EVMS_NAME_SIZE - 1);
		EngFncs->insert_thing(volume->subvolumes, subvol, INSERT_AFTER, NULL);
		LOG_DEBUG("Subvolume %s has handle %d\n", subvol_path, subvol->app_handle);
		free(subvol_path);
	}

	btrfs_util_destroy_subvolume_iterator(iter);

error:
	btrfs_umount(volume);
	LOG_EXIT_INT(rc);
	return rc;
}

static int btrfs_probe(logical_volume_t *volume)
{
	if ((volume->vol_size<<EVMS_VSECTOR_SIZE_SHIFT)<BTRFS_MIN_SIZE) return ENOENT;

	int rc=0;
	struct btrfs_super_block *sb_ptr;

	sb_ptr=EngFncs->engine_alloc(BTRFS_SIZE_OF_SUPER_BLOCK);
	if (sb_ptr)
	{
		rc = get_btrfs_super_block(volume, sb_ptr);
		if (rc==0 && strncmp((char*)&sb_ptr->magic, BTRFS_MAGIC, 8)==0)
		{
			volume->private_data=(void*)sb_ptr;
		}
		else
		{
			volume->private_data=NULL;
			EngFncs->engine_free(sb_ptr);
			rc = ENOENT;
		}
	}
	else rc = ENOMEM;

	if (!rc && !volume->subvolumes) {
		rc = btrfs_get_subvols(volume);
	}

	return rc;
};

static int btrfs_get_fs_size(logical_volume_t *volume, sector_count_t *size)
{
	int rc=EINVAL;
	struct btrfs_super_block *sb_ptr=(struct btrfs_super_block *)volume->private_data;

	if (!sb_ptr) return rc;

	if (strncmp((char*)&sb_ptr->magic, BTRFS_MAGIC, 8)==0)
	{
		*size = sb_ptr->total_bytes >> EVMS_VSECTOR_SIZE_SHIFT;
		rc = 0;
	}

	return rc;
};

static int btrfs_get_fs_limits(logical_volume_t *volume, sector_count_t *min_size, sector_count_t *max_volume_size, sector_count_t *max_object_size)
{
	int rc = EINVAL;
	struct btrfs_super_block *sb_ptr=(struct btrfs_super_block *)volume->private_data;

	if (!sb_ptr) return rc;

	if (strncmp((char*)&sb_ptr->magic, BTRFS_MAGIC, 8)==0)
	{
		if (sb_ptr->bytes_used>BTRFS_MIN_SIZE) *min_size=sb_ptr->bytes_used >> EVMS_VSECTOR_SIZE_SHIFT;
		else *min_size=BTRFS_MIN_SIZE >> EVMS_VSECTOR_SIZE_SHIFT;
		*max_volume_size = (sector_count_t) 1<<(64-EVMS_VSECTOR_SIZE_SHIFT);
		*max_object_size = (sector_count_t) 1<<(64-EVMS_VSECTOR_SIZE_SHIFT);
	}

	return rc;
};

static int btrfs_can_mkfs(logical_volume_t *volume)
{
	if (!have_mkfs_prog) return ENOSYS;

	int rc=0;
	if (EVMS_IS_MOUNTED(volume)) rc=EBUSY;
	else if ((volume->vol_size<<EVMS_VSECTOR_SIZE_SHIFT)<BTRFS_MIN_SIZE) rc=EPERM;
	return rc;
};

static int btrfs_can_unmkfs(logical_volume_t *volume)
{
	int  rc=0;
	if (EVMS_IS_MOUNTED(volume)) rc=EBUSY;
	return rc;
};

static int btrfs_can_fsck(logical_volume_t * UNUSED(volume))
{
	return ENOSYS;
};

static int btrfs_can_expand_by(logical_volume_t *volume, sector_count_t *delta_size)
{
	return ENOSYS;
};

static int btrfs_can_shrink_by(logical_volume_t *volume, sector_count_t *delta_size)
{
	return ENOSYS;
};

static int btrfs_mkfs_setup(logical_volume_t *volume, option_array_t *options)
{
	int rc = 0;
	LOG_ENTRY();

	volume->subvolumes = EngFncs->allocate_list();
	if (!volume->subvolumes)
	{
		rc = ENOMEM;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

static int btrfs_mkfs(logical_volume_t *volume, option_array_t *options)
{
	if (!have_mkfs_prog) return ENOSYS;

	int i;
	int opt=0;
	int status;
	int rc;
	char *cmd[5];
	cmd[opt++]=BTRFS_MKFS_PROG;

	for (i=0; i<options->count; i++)
	{
		if (!options->option[i].is_number_based)
		{
			if (!strcmp(options->option[i].name, MKFS_LABEL_NAME)) options->option[i].number=MKFS_LABEL_INDEX;
			else continue;
		}

		if (options->option[i].number==MKFS_LABEL_INDEX && options->option[i].value.s!=NULL)
		{
			cmd[opt++]="-L";
			cmd[opt++]=options->option[i].value.s;
		}
	}

	cmd[opt++]=EVMS_GET_DEVNAME(volume);
	cmd[opt++]=NULL;

	int fds[2];
	fds[0] = 0;
	fds[1] = open("/dev/null", O_WRONLY);

	pid_t pidm;
	pidm = EngFncs->fork_and_execvp(volume, cmd, NULL, fds, fds);
	if (pidm != -1)
	{
		waitpid(pidm, &status, 0);
		if (WIFEXITED(status)) rc=WEXITSTATUS(status);
		else rc=EINTR;
	}
	else rc=errno;

	if (rc==0) rc=btrfs_probe(volume);

	close(fds[1]);
	return rc;
};

static int btrfs_unmkfs(logical_volume_t *volume)
{
	if (EVMS_IS_MOUNTED(volume)) return EBUSY;
	if (!volume->private_data) return ENOMEM;

	int rc, fd;

	btrfs_do_destroy_subvols(volume);
	volume->subvolumes = NULL;

	fd = EngFncs->open_volume(volume, O_RDWR | O_EXCL, 0);
	if (fd<0) return EBUSY;

	memset((void *)volume->private_data, 0, BTRFS_SIZE_OF_SUPER_BLOCK);

	if (EngFncs->write_volume(volume, fd, volume->private_data, BTRFS_SIZE_OF_SUPER_BLOCK, BTRFS_SUPER_LOC)==BTRFS_SIZE_OF_SUPER_BLOCK) rc=0;
	else rc=EINVAL;

	EngFncs->engine_free(volume->private_data);
	volume->private_data=NULL;

	EngFncs->close_volume(volume, fd);

	return rc;
};

static int btrfs_expand(logical_volume_t *volume, sector_count_t *new_size)
{
	return ENOSYS;
};

static int btrfs_shrink(logical_volume_t *volume, sector_count_t requested_size, sector_count_t *new_size)
{
	return ENOSYS;
};

static int btrfs_discard(logical_volume_t *volume)
{
	if (volume->private_data)
	{
		EngFncs->engine_free(volume->private_data);
		volume->private_data = NULL;
	}
	return 0;
};

static int btrfs_get_option_count(task_context_t *context)
{
	int count;

	switch (context->action) {
	case EVMS_Task_mkfs:
		count = MKFS_OPTIONS_COUNT;
		break;
	case EVMS_Task_Create_Subvolume:
		count = CREATE_SUBVOL_OPTIONS_COUNT;
		break;
	default:
		count = -1;
		break;
	}

	return count;
};

static int init_mkfs_acceptable_objects(task_context_t * context)
{
	int rc;
	list_anchor_t global_volumes;
	list_element_t vol_list_iter;
	logical_volume_t * volume;

	rc = EngFncs->get_volume_list(NULL, NULL, 0, &global_volumes);
	if (!rc)
	{
		LIST_FOR_EACH(global_volumes, vol_list_iter, volume)
		{
			if ((volume->file_system_manager == NULL) && !EVMS_IS_MOUNTED(volume) && ((volume->vol_size<<EVMS_VSECTOR_SIZE_SHIFT)>BTRFS_MIN_SIZE) && have_mkfs_prog)
			{
				EngFncs->insert_thing(context->acceptable_objects, volume, INSERT_BEFORE, NULL);
			}
		}
		EngFncs->destroy_list(global_volumes);
	}

	return rc;
};

static int init_create_subvol_acceptable_objects(task_context_t *context)
{
        int rc;
        list_anchor_t global_volumes;
        list_element_t vol_list_iter;
        logical_volume_t * volume;

	rc = EngFncs->get_volume_list(NULL, NULL, 0, &global_volumes);
	if (!rc)
	{
		LIST_FOR_EACH(global_volumes, vol_list_iter, volume)
		{
			if (volume->file_system_manager && !strcmp(volume->file_system_manager->short_name, "BtrFS"))
			{
				EngFncs->insert_thing(context->acceptable_objects, volume, INSERT_BEFORE, NULL);
			}
		}
		EngFncs->destroy_list(global_volumes);
	}

	return rc;
};

static int btrfs_init_task(task_context_t *context)
{
	int rc=0;
	option_descriptor_t *opt;

	switch (context->action)
	{
		case EVMS_Task_mkfs:
			rc = init_mkfs_acceptable_objects(context);
			if (rc) return rc;
			context->option_descriptors->count = MKFS_OPTIONS_COUNT;

			opt = &context->option_descriptors->option[MKFS_LABEL_INDEX];
			opt->name = EngFncs->engine_strdup(MKFS_LABEL_NAME);
			opt->title = EngFncs->engine_strdup(_("Volume label"));
			opt->tip = EngFncs->engine_strdup(_("Set the label for the volume."));
			opt->type = EVMS_Type_String;
			opt->min_len = 1;
			opt->max_len = MAX_LABEL_LEN;
			opt->flags = EVMS_OPTION_FLAGS_NOT_REQUIRED;
			opt->value.s = EngFncs->engine_alloc(MAX_LABEL_LEN + 1);

			context->min_selected_objects = 1;
			context->max_selected_objects = 1;
			break;

		case EVMS_Task_Create_Subvolume:
			rc = init_create_subvol_acceptable_objects(context);
			if (rc) return rc;
			context->option_descriptors->count = CREATE_SUBVOL_OPTIONS_COUNT;

			opt = &context->option_descriptors->option[CREATE_SUBVOL_LABEL_INDEX];
			opt->name = EngFncs->engine_strdup(CREATE_SUBVOL_LABEL_NAME);
			opt->title = EngFncs->engine_strdup(_("Subvolume Path"));
			opt->tip = EngFncs->engine_strdup(_("Path for the new subvolume."));
			opt->type = EVMS_Type_String;
			opt->min_len = 1;
			opt->max_len = MAX_LABEL_LEN;
			opt->flags = EVMS_OPTION_FLAGS_NO_INITIAL_VALUE;
			opt->value.s = EngFncs->engine_alloc(MAX_LABEL_LEN + 1);

			context->min_selected_objects = 1;
			context->max_selected_objects = 1;
			break;

		default:
			rc = EINVAL;
			break;
	}

	return rc;
};

static int btrfs_set_option(task_context_t *context, u_int32_t index, value_t *value, task_effect_t *effect)
{
	int rc=0;

	if (!context || !value || !effect) return EFAULT;

	switch (context->action) {
	case EVMS_Task_mkfs:
		if (index == MKFS_LABEL_INDEX)
		{
			strncpy(context->option_descriptors->option[index].value.s, value->s, MAX_LABEL_LEN);
			if (strlen(value->s) > MAX_LABEL_LEN) MESSAGE(_("Volume label is truncated to \"%s\".\n"), context->option_descriptors->option[index].value.s);
		}
		break;
	case EVMS_Task_Create_Subvolume:
		if (index == CREATE_SUBVOL_LABEL_INDEX)
		{
			strncpy(context->option_descriptors->option[index].value.s, value->s, MAX_LABEL_LEN);
			if (strlen(value->s) > MAX_LABEL_LEN) MESSAGE(_("Subvolume label is truncated to \"%s\".\n"), context->option_descriptors->option[index].value.s);
		}
		break;
	default:
		rc = EINVAL;
		break;
	}

	return rc;
};

static int btrfs_set_volumes(task_context_t *context, list_anchor_t UNUSED(declined_volumes), task_effect_t * UNUSED(effect))
{
	int rc=0;

	if (context->action==EVMS_Task_mkfs)
	{
		logical_volume_t * vol;
		vol = EngFncs->first_thing(context->selected_objects, NULL);
		if (vol)
		{
			if (EVMS_IS_MOUNTED(vol)) rc=EBUSY;
			else if ((vol->vol_size<<EVMS_VSECTOR_SIZE_SHIFT)<BTRFS_MIN_SIZE) rc=EPERM;
			else if (!have_mkfs_prog) rc=ENOSYS;
		}
		else rc=ENODATA;
	}

	return rc;
};

static int btrfs_get_volume_info(logical_volume_t *volume, char *info_name, extended_info_array_t **info)
{
	return EINVAL;
};

static inline int btrfs_get_plugin_info(char *descriptor_name, extended_info_array_t **info)
{
	extended_info_array_t *Info;
	char version_string[64];
	char required_engine_api_version_string[64];
	char required_fsim_api_version_string[64];

	if (!info || descriptor_name) return EINVAL;

	*info = NULL;

	Info = EngFncs->engine_alloc(sizeof(extended_info_array_t) + 6 * sizeof(extended_info_t));
	if (!Info) return ENOMEM;

	Info->count = 6;

	sprintf(version_string, "%d.%d.%d", MAJOR_VERSION, MINOR_VERSION, PATCH_LEVEL);

	sprintf(required_engine_api_version_string, "%d.%d.%d", my_plugin_record->required_engine_api_version.major, my_plugin_record->required_engine_api_version.minor, my_plugin_record->required_engine_api_version.patchlevel);
	sprintf(required_fsim_api_version_string, "%d.%d.%d", my_plugin_record->required_plugin_api_version.fsim.major, my_plugin_record->required_plugin_api_version.fsim.minor, my_plugin_record->required_plugin_api_version.fsim.patchlevel);

	Info->info[0].name = EngFncs->engine_strdup("Short Name");
	Info->info[0].title = EngFncs->engine_strdup(_("Short Name"));
	Info->info[0].desc = EngFncs->engine_strdup(_("A short name given to this plug-in"));
	Info->info[0].type = EVMS_Type_String;
	Info->info[0].value.s = EngFncs->engine_strdup(my_plugin_record->short_name);

	Info->info[1].name = EngFncs->engine_strdup("Long Name");
	Info->info[1].title = EngFncs->engine_strdup(_("Long Name"));
	Info->info[1].desc = EngFncs->engine_strdup(_("A longer, more descriptive name for this plug-in"));
	Info->info[1].type = EVMS_Type_String;
	Info->info[1].value.s = EngFncs->engine_strdup(my_plugin_record->long_name);

	Info->info[2].name = EngFncs->engine_strdup("Type");
	Info->info[2].title = EngFncs->engine_strdup(_("Plug-in Type"));
	Info->info[2].desc = EngFncs->engine_strdup(_("There are various types of plug-ins, each responsible for some kind of storage object or logical volume."));
	Info->info[2].type = EVMS_Type_String;
	Info->info[2].value.s = EngFncs->engine_strdup(_("File System Interface Module"));

	Info->info[3].name = EngFncs->engine_strdup("Version");
	Info->info[3].title = EngFncs->engine_strdup(_("Plug-in Version"));
	Info->info[3].desc = EngFncs->engine_strdup(_("This is the version number of the plug-in."));
	Info->info[3].type = EVMS_Type_String;
	Info->info[3].value.s = EngFncs->engine_strdup(version_string);

	Info->info[4].name = EngFncs->engine_strdup("Required Engine Services Version");
	Info->info[4].title = EngFncs->engine_strdup(_("Required Engine Services Version"));
	Info->info[4].desc = EngFncs->engine_strdup(_("This is the version of the Engine services that this plug-in requires.  "
						      "It will not run on older versions of the Engine services."));
	Info->info[4].type = EVMS_Type_String;
	Info->info[4].value.s = EngFncs->engine_strdup(required_engine_api_version_string);

	Info->info[5].name = EngFncs->engine_strdup("Required Engine FSIM API Version");
	Info->info[5].title = EngFncs->engine_strdup(_("Required Engine FSIM API Version"));
	Info->info[5].desc = EngFncs->engine_strdup(_("This is the version of the Engine FSIM API that this plug-in requires.  "
						      "It will not run on older versions of the Engine FSIM API."));
	Info->info[5].type = EVMS_Type_String;
	Info->info[5].value.s = EngFncs->engine_strdup(required_fsim_api_version_string);

	*info = Info;

	return 0;
};

/*
 * https://gist.github.com/JonathonReinhart/8c0d90191c38af2dcadb102c4e202950
 */
static int mkdir_p(const char * path)
{
	int rc = 0;
	char * path_copy = NULL;
	char * p;

	LOG_ENTRY();

	LOG_DEBUG("Request to create dir %s\n", path);
	path_copy = strdup(path);
	if (!path_copy) {
		rc = ENOMEM;
		goto error;
	}

	for (p = path_copy + 1; *p; p++) {
		if (*p == '/') {
			/* Temporarily truncate */
			*p = '\0';

			LOG_DEBUG("Trying to create directory at %s\n", path_copy);
			if (mkdir(path_copy, S_IRWXU) != 0) {
				if (errno != EEXIST) {
					rc = errno;
					goto error;
				} else {
					LOG_DEBUG("Directory at %s exists, continuing\n", path_copy);
				}
			}
			*p = '/';
		}
	}

error:
	free(path_copy);
	LOG_EXIT_INT(rc);
	return rc;
}

static int btrfs_do_create_subvol(subvolume_t * subvol)
{
	int rc = 0;
	logical_volume_t * volume = subvol->volume;
	char * mount_path = "/mnt/destination/";
	char absolute_path[EVMS_NAME_SIZE + 18];

	LOG_ENTRY();
	if (!(subvol->flags & SUBVFLAG_CREATE) || (subvol->flags & SUBVFLAG_DESTROY)) {
		LOG_ERROR("%s subvolume flags are incorrect\n", subvol->name);
		rc = EINVAL;
		goto error;
	}

	if (!EVMS_IS_MOUNTED(volume)) {
		rc = btrfs_mount(volume, mount_path);
	}

	if (rc != 0)
	{
		goto error;
	}

	snprintf(absolute_path, EVMS_NAME_SIZE + 19, "%s%s", mount_path, subvol->short_name);

	if ((rc = mkdir_p(absolute_path))) {
		goto error;
	}

	enum btrfs_util_error btrfs_rc;
	btrfs_rc = btrfs_util_create_subvolume(absolute_path, 0, NULL, NULL);
	if (btrfs_rc != BTRFS_UTIL_OK)
	{
		LOG_ERROR("Error creating subvolume at %s (%s): %s\n", subvol->name, absolute_path, btrfs_util_strerror(btrfs_rc));
		rc = btrfs_rc;
		goto error;
	}

	subvol->flags = 0;
error:
	if (EVMS_IS_MOUNTED(volume)) {
		btrfs_umount(volume);
	}
	LOG_EXIT_INT(rc);
	return rc;
}

static int btrfs_do_destroy_subvol(subvolume_t * subvol)
{
	int rc = 0;
	logical_volume_t * volume = subvol->volume;
	char * mount_path = "/mnt/destination/";
	char absolute_path[EVMS_NAME_SIZE + 18];

	LOG_ENTRY();
	if (!(subvol->flags & SUBVFLAG_DESTROY) || (subvol->flags & SUBVFLAG_CREATE)) {
		LOG_ERROR("%s subvolume flags are incorrect\n", subvol->name);
		rc = EINVAL;
		goto error;
	}

	if (!EVMS_IS_MOUNTED(volume)) {
		rc = btrfs_mount(volume, mount_path);
	}

	if (rc != 0)
	{
		goto error;
	}

	snprintf(absolute_path, EVMS_NAME_SIZE + 19, "%s%s", mount_path, subvol->short_name);

	enum btrfs_util_error btrfs_rc;
	btrfs_rc = btrfs_util_delete_subvolume(absolute_path, 0);
	if (btrfs_rc != BTRFS_UTIL_OK)
	{
		LOG_ERROR("Error destroying subvolume at %s (%s): %s\n", subvol->name, absolute_path, btrfs_util_strerror(btrfs_rc));
		rc = btrfs_rc;
		goto error;
	}

	EngFncs->remove_thing(volume->subvolumes, subvol);
	EngFncs->free_subvolume(subvol);

error:
	if (EVMS_IS_MOUNTED(volume)) {
		btrfs_umount(volume);
	}
	LOG_EXIT_INT(rc);
	return rc;
}

static int btrfs_do_create_subvols(logical_volume_t * vol) {

        int rc = 0;
        list_element_t subiter1, subiter2;
        subvolume_t * subvol;

        LOG_ENTRY();

	EngFncs->sort_list(vol->subvolumes, sort_subvols_asc, NULL);

        LIST_FOR_EACH_SAFE(vol->subvolumes, subiter1, subiter2, subvol) {
                LOG_DEBUG("Creating subvolume %s\n", subvol->name);

                if (subvol->flags & SUBVFLAG_CREATE) {
                        rc = btrfs_do_create_subvol(subvol);
                }
        }

        LOG_EXIT_INT(rc);
        return rc;
}

static int btrfs_do_destroy_subvols(logical_volume_t * vol) {
        int rc = 0;
        list_element_t subiter1, subiter2;
        subvolume_t * subvol;

        LOG_ENTRY();

	EngFncs->sort_list(vol->subvolumes, sort_subvols_desc, NULL);

        LIST_FOR_EACH_SAFE(vol->subvolumes, subiter1, subiter2, subvol) {
                LOG_DEBUG("Destroying subvolume %s\n", subvol->name);

                if (subvol->flags & SUBVFLAG_DESTROY) {
                        rc = btrfs_do_destroy_subvol(subvol);
                }
        }

        LOG_EXIT_INT(rc);
        return rc;

}

subvolume_t * btrfs_find_subvol(logical_volume_t *volume, char * name)
{
	list_element_t iter;
	subvolume_t * subvol;

	LIST_FOR_EACH(volume->subvolumes, iter, subvol) {
		if (subvol && !strcmp(subvol->name, name)) {
			return subvol;
		}
	}

	return NULL;
}

static int btrfs_create_subvol(logical_volume_t *volume, option_array_t * options, list_anchor_t output_objects)
{
	int rc = 0;
	int i;
	char * path;
	subvolume_t * subvol;
	char full_path[EVMS_NAME_SIZE+1];

	LOG_ENTRY();
	for (i = 0; i < options->count; i++)
	{
		if (!options->option[i].is_number_based)
		{
			if (!strcmp(options->option[i].name, CREATE_SUBVOL_LABEL_NAME)) options->option[i].number = CREATE_SUBVOL_LABEL_INDEX;
			else continue;
		}

		if (options->option[i].number == CREATE_SUBVOL_LABEL_INDEX && options->option[i].value.s != NULL)
		{
			path = options->option[i].value.s;
		}
	}

	if (!path) {
		LOG_ERROR("Subvolume path is not defined\n");
		rc = EINVAL;
		goto error;
	}

	LOG_DEBUG("Path was %s\n", path);
	// Trim slashes from end of string
	while (path[strlen(path) - 1] == '/')
	{
		path[strlen(path) - 1] = '\0';
		LOG_EVERYTHING("Trim from end: %s\n", path);

		if (strlen(path) == 0) {
			rc = EINVAL;
			goto error;
		}
	}

	// Trim slashes from start of string
	while (path[0] == '/')
	{
		path++;
		LOG_EVERYTHING("Trim from beginning: %s\n", path);
	}

	LOG_DEBUG("Path (trimmed) is %s\n", path);

	subvol_name_to_btrfs_name(path, volume->name, full_path);

	subvol = btrfs_find_subvol(volume, full_path);

	if (subvol) {
		/* we've found a subvolume with given name */
		if (subvol->flags & SUBVFLAG_DESTROY) {
			LOG_DEBUG("Destroying subvolume %s of volume %s is cancelled\n", subvol->name, subvol->volume->name);
			subvol->flags &= ~SUBVFLAG_DESTROY;
		}
	} else {
		if (EngFncs->allocate_subvolume(full_path, &subvol)) {
			rc = ENOMEM;
			goto error;
		}

		subvol->volume = volume;
		strncpy(subvol->short_name, path, EVMS_NAME_SIZE - 1);
		EngFncs->insert_thing(volume->subvolumes, subvol, INSERT_AFTER, NULL);
		subvol->flags = SUBVFLAG_CREATE;
		LOG_DEBUG("Subvolume %s has handle %d\n", full_path, subvol->app_handle);
		EngFncs->insert_thing(output_objects, subvol, INSERT_AFTER, NULL);
	}

error:
	LOG_EXIT_INT(rc);
	return rc;
}

static int btrfs_can_destroy_subvol(subvolume_t * subvol)
{
	// TODO check if there are children subvolumes
	list_anchor_t subvols = subvol->volume->subvolumes;
	list_element_t iter;
	subvolume_t * sample;
	char subvol_path[EVMS_NAME_SIZE + 2];
	char sample_path[EVMS_NAME_SIZE + 2];

	snprintf(subvol_path, EVMS_NAME_SIZE + 2, "%s/", subvol->name);

	LIST_FOR_EACH(subvols, iter, sample) {

		if (sample->flags & SUBVFLAG_DESTROY) {
			LOG_DEBUG("%s subvol marked for destroy, skipping\n", sample->name);
			continue;
		}

		snprintf(sample_path, EVMS_NAME_SIZE + 2, "%s/", sample->name);

		unsigned int rc;
		rc = strcmp(subvol_path, sample_path);

		if (rc) {
			if (strstr(sample_path, subvol_path)) {
				LOG_DEBUG("%s is a substring to %s\n", subvol_path, sample_path);
				return EBUSY;
			}
		} else {
			LOG_DEBUG("%s is equal to %s, skipping\n", subvol_path, sample_path);
			continue;
		}

		LOG_DEBUG("Comparing strings %s and %s gave %d result\n", subvol_path, sample_path, rc);
	}

	return 0;
}

static int btrfs_destroy_subvol(subvolume_t * subvol)
{
	int rc = 0;
	logical_volume_t * volume = subvol->volume;
	LOG_ENTRY();
	LOG_DEBUG("Destroying subvolume at path %s\n", subvol->name);
	if (subvol->flags & SUBVFLAG_CREATE) {
		/* Not created yet; safe to destroy subvol */

		EngFncs->remove_thing(volume->subvolumes, subvol);
		EngFncs->free_subvolume(subvol);
	} else {
		subvol->flags = SUBVFLAG_DESTROY;
	}

	LOG_EXIT_INT(rc);
	return rc;
}

static fsim_functions_t fsim_ops =
{
	.setup_evms_plugin	= btrfs_setup,
	.cleanup_evms_plugin	= btrfs_cleanup,
	.probe			= btrfs_probe,
	.get_fs_size		= btrfs_get_fs_size,
	.get_fs_limits		= btrfs_get_fs_limits,
	.mkfs			= btrfs_mkfs,
	.mkfs_setup		= btrfs_mkfs_setup,
	.can_mkfs		= btrfs_can_mkfs,
	.can_unmkfs		= btrfs_can_unmkfs,
	.can_fsck		= btrfs_can_fsck,
	.can_expand_by		= btrfs_can_expand_by,
	.can_shrink_by		= btrfs_can_shrink_by,
	.discard		= btrfs_discard,
	.unmkfs			= btrfs_unmkfs,
	.expand			= btrfs_expand,
	.shrink			= btrfs_shrink,
	.get_option_count	= btrfs_get_option_count,
	.init_task		= btrfs_init_task,
	.set_option		= btrfs_set_option,
	.set_volumes		= btrfs_set_volumes,
	.get_volume_info	= btrfs_get_volume_info,
	.get_plugin_info	= btrfs_get_plugin_info,
	.create_subvol		= btrfs_create_subvol,
	.can_destroy_subvol	= btrfs_can_destroy_subvol,
	.destroy_subvol		= btrfs_destroy_subvol,
	.do_create_subvols	= btrfs_do_create_subvols,
	.do_destroy_subvols	= btrfs_do_destroy_subvols,
};

plugin_record_t btrfs_plugrec =
{
	.id = EVMS_BTRFS_PLUGIN_ID,
	.version = {
		.major		= MAJOR_VERSION,
		.minor		= MINOR_VERSION,
		.patchlevel	= PATCH_LEVEL
	},
	.required_engine_api_version = {
		.major		= 15,
		.minor		= 0,
		.patchlevel	= 0
	},
	.required_plugin_api_version = {
		.fsim = {
			.major		= 11,
			.minor		= 0,
			.patchlevel	= 0
		}
	},
	.short_name = EVMS_BTRFS_PLUGIN_SHORT_NAME,
	.long_name = EVMS_BTRFS_PLUGIN_LONG_NAME,
	.oem_name = EVMS_ALT_LINUX_OEM_NAME,
	.functions = {
		.fsim = &fsim_ops
	},
	.container_functions = NULL,
};

plugin_record_t *evms_plugin_records[] =
{
	&btrfs_plugrec,
	NULL
};
