#define _GNU_SOURCE

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "imsm.h"
#include "helpers.h"
#include "log_utils.h"

controller_info_t* controllers = NULL;
controller_info_t* controllers_tail = NULL;

#ifndef MDADM_PROGRAM
#define MDADM_PROGRAM "/sbin/mdadm"
#endif

typedef void (*whitespace_split_handler)(const char* part, size_t len, void* userdata);
static void whitespace_split(const char* str, size_t len, void* userdata, whitespace_split_handler handler) {
    if (len == 0)
        return;

    size_t i = 0;
    while (i < len) {
        while (i < len && isspace(str[i]))
            ++i;

        const char* start = str + i;
        while (i < len && isalnum(str[i]))
            ++i;
        const char* end = str + i;

        if (start != end)
            handler(start, end - start, userdata);
    }
}

/**
 * @brief Handler for whitespace_split() for raid level extraction
 *
 * @param part - start of the string part
 * @param len - length of the string part
 * @param capabilities - output raid level capabilities
 */
static void handle_raid_level(const char* part, size_t len, void* capabilities) {
    controller_cap_e* caps = capabilities;

    if (len <= 4 || strncmp("raid", part, 4) != 0)
        return;

    part += 4;
    len -= 4;

    if (len == 1) {
        switch(part[0]) {
            case '0':
                *caps |= controller_cap_raid0;
                return;
            case '1':
                *caps |= controller_cap_raid1;
                return;
            case '5':
                *caps |= controller_cap_raid5;
                return;
        }
    }
    else if (len == 2 && part[0] == '1' && part[1] == '0')
        *caps |= controller_cap_raid10;
}

/**
 * @brief Parse the raid level capabilities from "RAID Levels" mdadm's output
 *
 * @param line - string ended with '\n' or '\0'
 *
 * @return raid level capabilities
 */
static controller_cap_e parse_supported_raid_levels(const char* line) {
    LOG_ENTRY();

    const char* end = line;
    while (*end != '\0' && *end != '\n')
        ++end;

    controller_cap_e caps = 0;
    whitespace_split(line, end - line, &caps, handle_raid_level);

    return LOG_EXIT(caps);
}

/**
 * @brief Parse controller capabilities from mdadm's output
 *
 * @param mdadm_output - mdadm's output buffer
 * @param capabilities - output capabilities
 *
 * @return 0 on success, otherwise error code
 */
static int parse_capabilities(const char* mdadm_output, controller_cap_e* capabilities) {
    LOG_ENTRY();

    char* pos = strstr(mdadm_output, "RAID Levels :");
    if (pos == NULL)
        return LOG_EXIT(EINVAL);

    pos += sizeof("RAID Levels :") - 1;
    *capabilities = parse_supported_raid_levels(pos);

    return LOG_EXIT(0);
}

/**
 * @brief Extract the controller capabilities from mdadm --detail-platform
 *
 * @param controller_path - path to the controller
 * @param capabilities - output capabilities
 *
 * @return 0 on success, otherwise error code
 */
static int get_controller_capabilities(const char* controller_path, controller_cap_e* capabilities) {
    LOG_ENTRY();

    char* args[5] = {
        MDADM_PROGRAM,
        "--detail-platform",
        "--metadata=imsm",
        (char*)controller_path,
        NULL,
    };

    autocleanup(char*) buf;
    int rc = fork_and_execvp(args, &buf);
    if (rc != 0)
        return LOG_EXIT(rc);

    rc = parse_capabilities(buf, capabilities);
    return LOG_EXIT(rc);
}

/**
 * @brief Recursively search controller from device path, transfrom the input path
 * to controller path by adding '\0' in right place, and extract controller
 * capabilities
 *
 * @param path - input device path
 * @param capabilities - output capabilities
 *
 * @return 0 on success, otherwise error code
 */
static int extract_controller_info_recursive(char* path, controller_cap_e* capabilities) {
    LOG_ENTRY();

    if (strcmp(path, "/sys/devices") == 0) {
        return LOG_EXIT(EINVAL);
    }

    char* pos = strrchr(path, '/');
    if (pos == NULL)
        return LOG_EXIT(EINVAL);

    *pos = '\0';

    int res = extract_controller_info_recursive(path, capabilities);
    if (res == EINVAL) {
        *pos = '/';

        controller_cap_e caps = 0;
        int rc = get_controller_capabilities(path, &caps);
        if (rc == 0) {
            *capabilities = caps;
            return LOG_EXIT(0);
        }
    }

    return LOG_EXIT(res);
}

/**
 * @brief Get controller path and extract controller capabilities by
 * disk path
 *
 * @param disk_sysfs_path - the disk path in sysfs (will be mutated)
 * @param capabilities - output capabilities
 *
 * @return controller path (in disk_sysfs_path memory) or NULL in not found
 */
static char* extract_controller_info(char* disk_sysfs_path, controller_cap_e* capabilities) {
    LOG_ENTRY();

    int rc = extract_controller_info_recursive(disk_sysfs_path, capabilities);
    if (rc == 0)
        return LOG_EXIT(disk_sysfs_path);

    return LOG_EXIT((void*)NULL);
}

/**
 * @brief Get controller info from disk path
 *
 * @param disk_sysfs_path - the disk path in sysfs
 *
 * @return controller info or NULL if not found
 */
static controller_info_t* get_controller_from_cache(char* disk_sysfs_path) {
    LOG_ENTRY();

    size_t disk_path_len = strlen(disk_sysfs_path);

    FOREACH_LIST(controller, controllers) {
        size_t controller_path_len = strlen(controller->sysfs_path);
        if (controller_path_len < disk_path_len &&
            strncmp(controller->sysfs_path, disk_sysfs_path, controller_path_len) == 0)
            return LOG_EXIT((void*)controller);
    }

    return LOG_EXIT((void*)NULL);
}

/**
 * @brief Get controller info from disk's storage object
 *
 * @param disk - disk's storage object
 *
 * @return controller info or NULL if not found
 */
controller_info_t* get_controller(storage_object_t* disk) {
    LOG_ENTRY();

    char sysfs_block_path[256 + 1] = {0};
    snprintf(sysfs_block_path, sizeof(sysfs_block_path) - 1, "/sys/class/block/%s", disk->name);

    char sysfs_link[4096 + 1] = {0};
    ssize_t sz = readlink(sysfs_block_path, sysfs_link, sizeof(sysfs_link) - 1);
    if (sz < 0)
        return LOG_EXIT((void*)NULL);

    if (strncmp(sysfs_link, "../../", sizeof("../../") - 1) != 0)
        return LOG_EXIT((void*)NULL);

    char* sysfs_path = sysfs_link + 1;
    memcpy(sysfs_path, "/sys", sizeof("/sys") - 1);

    controller_info_t* ctrl = get_controller_from_cache(sysfs_path);
    if (ctrl)
        return LOG_EXIT((void*)ctrl);

    /* Trying to get info about controller */
    controller_cap_e caps = 0;
    char* controller_path = extract_controller_info(sysfs_path, &caps);

    if (!controller_path)
        return LOG_EXIT((void*)NULL);

    ctrl = EngFncs->engine_alloc(sizeof(controller_info_t));
    strncpy(ctrl->sysfs_path, controller_path, sizeof(ctrl->sysfs_path) - 1);
    ctrl->capabilities = caps;
    ctrl->next_tied = NULL;

    if (controllers_tail) {
        controllers_tail->next_tied = ctrl;
        controllers_tail = ctrl;
    }
    else {
        controllers = ctrl;
        controllers_tail = ctrl;
    }

    return ctrl;
}

