#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <sys/wait.h>
#include <unistd.h>

#include "log_utils.h"

static char* devid2kname(dev_t devid) {
    char path[30];
    char link[PATH_MAX];
    static char devnm[32];
    char* cp;
    int n;

    /* Look at the
     * /sys/dev/block/%d:%d link which must look like
     * and take the last component.
     */
    sprintf(path, "/sys/dev/block/%d:%d", major(devid), minor(devid));
    n = readlink(path, link, sizeof(link) - 1);
    if (n > 0) {
        link[n] = 0;
        cp = strrchr(link, '/');
        if (cp) {
            strcpy(devnm, cp + 1);
            return devnm;
        }
    }
    return NULL;
}

static char* stat2kname(struct stat* st) {
    if ((S_IFMT & st->st_mode) != S_IFBLK)
        return NULL;

    return devid2kname(st->st_rdev);
}

static char* fd2kname(int fd) {
    struct stat stb;

    if (fstat(fd, &stb) == 0)
        return stat2kname(&stb);

    return NULL;
}

#define NVME_SUBSYS_PATH "/sys/devices/virtual/nvme-subsystem/"

/* Check whether the nvme device is represented by nvme subsytem,
 * if yes virtual path should be changed to hardware device path,
 * to allow IMSM capabilities detection.
 * Returns:
 *	hardware path to device - if the device is represented via
 *		nvme virtual subsytem
 *	NULL - if the device is not represented via nvme virtual subsytem
 */
static char* get_nvme_multipath_dev_hw_path(const char* dev_path) {
    DIR* dir;
    struct dirent* ent;
    char* rp = NULL;

    if (strncmp(dev_path, NVME_SUBSYS_PATH, strlen(NVME_SUBSYS_PATH)) != 0)
        return NULL;

    dir = opendir(dev_path);
    if (!dir)
        return NULL;

    for (ent = readdir(dir); ent; ent = readdir(dir)) {
        char buf[strlen(dev_path) + strlen(ent->d_name) + 1];

        /* Check if dir is a controller, ignore namespaces*/
        if (!(strncmp(ent->d_name, "nvme", 4) == 0) || (strrchr(ent->d_name, 'n') != &ent->d_name[0]))
            continue;

        sprintf(buf, "%s/%s", dev_path, ent->d_name);
        rp = realpath(buf, NULL);
        break;
    }

    closedir(dir);
    return rp;
}

/* Description: Return part or whole realpath for the dev
 * Parameters:
 *	dev - the device to be quered
 *	dev_level - level of "/device" entries. It allows to caller to access
 *		    virtual or physical devices which are on "path" to quered
 *		    one.
 *	buf - optional, must be PATH_MAX size. If set, then will be used.
 */
static char* devt_to_devpath(dev_t dev, int dev_level, char* buf) {
    char device[PATH_MAX];
    char* hw_path;
    int i;
    unsigned long device_free_len = sizeof(device) - 1;
    char dev_str[] = "/device";
    unsigned long dev_str_len = strlen(dev_str);

    snprintf(device, sizeof(device), "/sys/dev/block/%d:%d", major(dev), minor(dev));

    /* If caller wants block device, return path to it even if it is exposed
     * via virtual layer.
     */
    if (dev_level == 0)
        return realpath(device, buf);

    device_free_len -= strlen(device);
    for (i = 0; i < dev_level; i++) {
        if (device_free_len < dev_str_len)
            return NULL;

        strncat(device, dev_str, device_free_len);

        /* Resolve nvme-subsystem abstraction if needed
         */
        device_free_len -= dev_str_len;
        if (i == 0) {
            char rp[PATH_MAX];

            if (!realpath(device, rp))
                return NULL;
            hw_path = get_nvme_multipath_dev_hw_path(rp);
            if (hw_path) {
                strcpy(device, hw_path);
                device_free_len = sizeof(device) - strlen(device) - 1;
                free(hw_path);
            }
        }
    }

    return realpath(device, buf);
}

static char* diskfd_to_devpath(int fd, int dev_level, char* buf) {
    /* return the device path for a disk, return NULL on error or fd
     * refers to a partition
     */
    struct stat st;

    if (fstat(fd, &st) != 0)
        return NULL;
    if (!S_ISBLK(st.st_mode))
        return NULL;

    return devt_to_devpath(st.st_rdev, dev_level, buf);
}

static int load_sys(char* path, char* buf, int len) {
    int fd = open(path, O_RDONLY);
    int n;
    if (fd < 0)
        return -1;
    n = read(fd, buf, len);
    close(fd);
    if (n < 0 || n >= len)
        return -1;
    buf[n] = 0;
    if (n && buf[n - 1] == '\n')
        buf[n - 1] = 0;
    return 0;
}

/* Description: Read text value of dev_path/entry field
 * Parameters:
 *	dev_path - sysfs path to the device
 *	entry - entry to be read
 *	buf - buffer for read value
 *	len - size of buf
 *	verbose - error logging level
 */
static int devpath_to_char(const char* dev_path, const char* entry, char* buf, int len, int verbose) {
    char path[PATH_MAX];

    snprintf(path, sizeof(path), "%s/%s", dev_path, entry);
    if (load_sys(path, buf, len)) {
        if (verbose)
            fprintf(stderr, "Cannot read %s, aborting\n", path);
        return 1;
    }

    return 0;
}

static int nvme_get_serial(int fd, void* buf, size_t buf_len) {
    char path[PATH_MAX];
    char* name = fd2kname(fd);

    if (!name)
        return 1;

    if (strncmp(name, "nvme", 4) != 0)
        return 1;

    if (!diskfd_to_devpath(fd, 1, path))
        return 1;

    return devpath_to_char(path, "serial", buf, buf_len, 0);
}

static int scsi_get_serial(int fd, void* buf, size_t buf_len) {
    unsigned char rsp_buf[255];
    unsigned char inq_cmd[] = {INQUIRY, 1, 0x80, 0, sizeof(rsp_buf), 0};
    unsigned char sense[32];
    struct sg_io_hdr io_hdr;
    int rv;
    unsigned int rsp_len;

    memset(&io_hdr, 0, sizeof(io_hdr));
    io_hdr.interface_id = 'S';
    io_hdr.cmdp = inq_cmd;
    io_hdr.cmd_len = sizeof(inq_cmd);
    io_hdr.dxferp = rsp_buf;
    io_hdr.dxfer_len = sizeof(rsp_buf);
    io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
    io_hdr.sbp = sense;
    io_hdr.mx_sb_len = sizeof(sense);
    io_hdr.timeout = 5000;

    rv = ioctl(fd, SG_IO, &io_hdr);

    if (rv)
        return rv;

    if ((io_hdr.info & SG_INFO_OK_MASK) != SG_INFO_OK)
        return -1;

    rsp_len = rsp_buf[3];

    if (!rsp_len || buf_len < rsp_len)
        return -1;

    memcpy(buf, &rsp_buf[4], rsp_len);

    return 0;
}

int imsm_read_serial(int fd, char* devname, u_int8_t* serial, size_t serial_buf_len) {
    char buf[50];
    int rv;
    size_t len;
    char* dest;
    char* src;
    unsigned int i;

    memset(buf, 0, sizeof(buf));

    rv = nvme_get_serial(fd, buf, sizeof(buf));

    if (rv)
        rv = scsi_get_serial(fd, buf, sizeof(buf));

    if (rv != 0) {
        if (devname)
            fprintf(stderr, "Failed to retrieve serial for %s\n", devname);
        return rv;
    }

    /* trim all whitespace and non-printable characters and convert
     * ':' to ';'
     */
    for (i = 0, dest = buf; i < sizeof(buf) && buf[i]; i++) {
        src = &buf[i];
        if (*src > 0x20) {
            /* ':' is reserved for use in placeholder serial
             * numbers for missing disks
             */
            if (*src == ':')
                *dest++ = ';';
            else
                *dest++ = *src;
        }
    }
    len = dest - buf;
    dest = buf;

    if (len > serial_buf_len) {
        /* truncate leading characters */
        dest += len - serial_buf_len;
        len = serial_buf_len;
    }

    memset(serial, 0, serial_buf_len);
    memcpy(serial, dest, len);

    return 0;
}

int fork_and_execvp(char** argv, char** output_buffer) {
    LOG_ENTRY();

    LOG_DETAILS("arguments:\n");
    for (char** argp = argv; *argp; ++argp) {
        LOG_DETAILS("arg[%li]: %s\n", argp - argv, *argp);
    }

    int rc = 0;
    int fds[2];
    if (output_buffer) {
        *output_buffer = NULL;
        if ((*output_buffer = EngFncs->engine_alloc(MAX_USER_MESSAGE_LEN)) == NULL) {
            rc = ENOMEM;
            goto fail;
        }

        if (pipe(fds) < 0) {
            rc = ENOSYS;
            goto fail;
        }

        fcntl(fds[0], F_SETFL, fcntl(fds[0], F_GETFL, 0) | O_NONBLOCK);
        fcntl(fds[1], F_SETFL, fcntl(fds[1], F_GETFL, 0) | O_NONBLOCK);
    }

    pid_t pidm = EngFncs->fork_and_execvp(NULL, argv, NULL, output_buffer ? fds : NULL, output_buffer ? fds : NULL);
    if (pidm != -1) {
        int status;
        waitpid(pidm, &status, 0);
        if (WIFEXITED(status)) {
            if (output_buffer)
                read(fds[0], *output_buffer, MAX_USER_MESSAGE_LEN);
            rc = WEXITSTATUS(status);
            LOG_DETAILS("%s completed with exit code %d \n", argv[0], rc);
        }
        else if (WIFSIGNALED(status)) {
            LOG_DETAILS("%s killed by signal %d\n", argv[0], WTERMSIG(status));
            rc = EINTR;
        }
        else if (WIFSTOPPED(status)) {
            LOG_DETAILS("%s stopped by signal %d\n", argv[0], WSTOPSIG(status));
            rc = EINTR;
        }
        else {
            if (WIFCONTINUED(status)) {
                LOG_DETAILS("%s continued\n", argv[0]);
            }
            /*
             * The process didn't exit. It must have been
             * interrupted by a signal.
             */
            rc = EINTR;
        }
    }
    else {
        LOG_SERIOUS("Failed to fork and exec %s.  Error code is %d: %s\n", argv[0], rc, EngFncs->strerror(rc));
        rc = errno;
    }

    if (output_buffer) {
        close(fds[0]);
        close(fds[1]);
    }

    return LOG_EXIT(rc);

fail:
    if (output_buffer) {
        EngFncs->engine_free(*output_buffer);
        *output_buffer = NULL;
    }
    return LOG_EXIT(rc);
}

char* strdup_cat(const char* a, const char* b) {
    size_t la = strlen(a);
    size_t lb = strlen(b);

    char* result = EngFncs->engine_alloc(la + lb + 1);
    result[la + lb] = '\0';
    memcpy(result, a, la);
    memcpy(result + la, b, lb);

    return result;
}

void plugin_free(void* a) {
    EngFncs->engine_free(*((void**)a));
}
