#define _GNU_SOURCE

#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <plugin.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <unistd.h>

#include "sb.h"
#include "imsm.h"
#include "helpers.h"
#include "log_utils.h"
#include "info.h"
#include "controller.h"

#define MD_MAJOR 9
#define CONTAINER_PREFIX "imsm"

engine_functions_t* EngFncs;

disk_infos_store_t disk_infos = {{}, 0};
container_infos_store_t container_infos = {{}, 0};
volume_infos_store_t volume_infos = {{}, 0};

/* Read the whole file into sector_count_t integer */
static sector_count_t read_sector_count(const char* offset_path) {
    LOG_ENTRY();

    int fd = open(offset_path, O_RDONLY);
    if (fd < 0)
        return LOG_EXIT(~(sector_count_t)0);

    char buf[256];
    read(fd, buf, sizeof(buf) - 1);
    close(fd);

    sector_count_t result = atoll(buf);
    return LOG_EXIT(result);
}

/* Returns true if vol and dev_path represent the same md device */
static boolean is_our_mddev(volume_info_t* vol, const char* dev_path, const char* disk_name) {
    LOG_ENTRY();

    /* Compare offsets from superblock and sysfs.
     * If offsets match, "vol" and "dev_path" represent the same md device.
     */

    imsm_map_t* map = get_imsm_map(vol->dev, MAP_0);
    sector_count_t offset = get_imsm_map_pba_of_lba0(map);

    char offset_path[1024 + 1];
    snprintf(offset_path, sizeof(offset_path) - 1, "%s/md/dev-%s/offset", dev_path, disk_name);
    sector_count_t sysfs_offset = read_sector_count(offset_path);

    boolean result = offset == sysfs_offset;
    return LOG_EXIT(result);
}

/**
 * @brief Try find prettified md device symlink from "/dev/md/"
 *
 * @param dev_name - device name (md[0-9]+)
 *
 * @return newly allocated string with relative path (pwd=/dev/) to symlink if it exists, otherwise returns NULL
 */
static char* find_prettified_symlink_mddev_name(const char* dev_name) {
    struct dirent *dir;
    DIR* d = opendir("/dev/md/");
    if (!d)
        return LOG_EXIT((void*)NULL);

    while ((dir = readdir(d)) != NULL) {
        if (dir->d_type != DT_LNK)
            continue;

        autocleanup(char*) path = strdup_cat("/dev/md/", dir->d_name);
        char sl_path[256 + 1];
        ssize_t sz = readlink(path, sl_path, sizeof(sl_path) - 1);
        if (sz < 0)
            continue;

        sl_path[sz] = '\0';
        autocleanup(char*) sl_full_path = strdup_cat("/dev/md/", sl_path);

        char* dev_path = realpath(sl_full_path, NULL);
        if (strncmp(dev_path, "/dev/", 5) == 0 && strcmp(dev_path + 5, dev_name) == 0) {
            char* result = strdup_cat("md/", dir->d_name);
            free(dev_path);
            closedir(d);
            return LOG_EXIT((void*)result);
        }
    }

    closedir(d);
    return LOG_EXIT((void*)NULL);
}

/* Makes evms volume name from imsm_dev_t::volume field */
static char* make_evms_volume_name(char* destination, size_t max_len, volume_info_t* vol) {
    LOG_ENTRY();

    size_t container_index = vol->container - container_infos.data;

    char imsm_name[sizeof(vol->dev->volume) + 1];
    imsm_name[sizeof(imsm_name) - 1] = '\0';
    memcpy(imsm_name, vol->dev->volume, sizeof(vol->dev->volume));

    snprintf(destination, max_len, "imsm/%s_%s_%zu", vol->container->evms_data->name, imsm_name, container_index);

    return LOG_EXIT(destination);
}

/* Returns a newly allocated string with /dev/md${minor} path */
static char* make_mddev_path(int minor) {
    char minor_str[32];
    snprintf(minor_str, sizeof(minor_str) - 1, "%i", minor);
    return strdup_cat("/dev/md", minor_str);
}

/* Returns a newly allocated string with ../md${minor} relative path (pwd=/dev/md/) */
static char* make_mddev_relative_path(int minor) {
    char minor_str[32];
    snprintf(minor_str, sizeof(minor_str) - 1, "%i", minor);
    return strdup_cat("../md", minor_str);
}

/* Get md device minor number */
static int get_mddev_minor(const char* name) {
    LOG_ENTRY();

    autocleanup(char*) dev_path = strdup_cat("/dev/", name);
    struct stat st;
    memset(&st, 0, sizeof(st));

    stat(dev_path, &st);
    int min = minor(st.st_rdev);
    return LOG_EXIT(min);
}

/**
 * @brief Creates symlink in /dev/md/ if it not exists
 *
 * @param link_path - link path
 * @param md_minor - md device minor number
 *
 * @return 0 if symlink exists or was created, otherwise any non-zero error code
 */
static int ensure_md_symlink(const char* link_path, int md_minor) {
    LOG_ENTRY();

    int rc = mkdir("/dev/imsm", 0755);
    if (rc == -1 && errno != EEXIST)
        return LOG_EXIT(errno);

    autocleanup(char*) md_relative_path = make_mddev_relative_path(md_minor);
    rc = symlink(md_relative_path, link_path);
    if (rc == -1 && errno == EEXIST)
        rc = 0;

    return LOG_EXIT(rc);
}

/**
 * @brief Try find relative path (pwd=/dev) of already assembled imsm volume
 *
 * @param vol - volume to search
 * @param evms_container - conuming evms container
 *
 * @return newly allocated string with relative device path if volume is assembled, otherwise NULL
 */
static char* find_usable_mddev_name(volume_info_t* vol, storage_container_t* evms_container) {
    LOG_ENTRY();

    list_element_t iter;
    storage_object_t* evms_disk = EngFncs->first_thing(evms_container->objects_consumed, &iter);

    char holders_dir[256 + 1] = "/sys/class/block/";
    strncat(holders_dir, evms_disk->name, sizeof(holders_dir) - 1);
    strncat(holders_dir, "/holders/", sizeof(holders_dir) - 1);

    struct dirent *dir;
    DIR* d = opendir(holders_dir);
    if (!d)
        return LOG_EXIT((void*)NULL);

    /* Foreach dev in /sys/class/block/${evms_disk->name}/holders/ */
    while ((dir = readdir(d)) != NULL) {
        if (dir->d_name[0] == '.')
            continue;

        autocleanup(char*) dev_path = strdup_cat(holders_dir, dir->d_name);

        if (is_our_mddev(vol, dev_path, evms_disk->name)) {
            char* result = find_prettified_symlink_mddev_name(dir->d_name);

            /* Try create symlink with prettified name */
            if (!result) {
                const size_t max_len = 256;
                if ((result = EngFncs->engine_alloc(max_len + 1))) {
                    make_evms_volume_name(result, max_len, vol);

                    autocleanup(char*) link_path = strdup_cat("/dev/", result);
                    int md_minor = get_mddev_minor(dev_path);

                    if (ensure_md_symlink(link_path, md_minor) != 0) {
                        EngFncs->engine_free(result);
                        result = NULL;
                    }
                }
            }

            if (!result) {
                /* If all fails, use standard mddev name */
                result = EngFncs->engine_alloc(strlen(dir->d_name) + 1);
                strcpy(result, dir->d_name);
            }

            closedir(d);
            return result;
        }
    }
    closedir(d);

    return LOG_EXIT((void*)NULL);
}

/* Allocates a new string with /dev/${name} path */
static char* make_dev_name(const char* name) {
    char* result = EngFncs->engine_alloc(sizeof("/dev/") + sizeof(((storage_object_t*)NULL)->name));
    strcpy(result, "/dev/");
    strcat(result, name);
    return result;
}

/* Returns free md device minor number */
static int find_free_md_minor() {
    LOG_ENTRY();

    const size_t minor_table_len = 256;
    boolean hold_table[minor_table_len];
    memset(hold_table, 0, sizeof(hold_table));

    list_anchor_t objects;
    int rc = EngFncs->get_object_list(DISK | SEGMENT | REGION, DATA_TYPE, NULL, NULL, VALID_INPUT_OBJECT, &objects);

    if (rc == 0) {
        /* Avoid minor collision with other evms objects (plugins/md) */
        list_element_t iter;
        storage_object_t* obj;
        LIST_FOR_EACH(objects, iter, obj) {
            if (obj->dev_major == MD_MAJOR && obj->dev_minor >= 0 && obj->dev_minor < minor_table_len) {
                hold_table[obj->dev_minor] = true;
                LOG_DETAILS("md minor %i is occupied by object %s\n", obj->dev_minor, obj->name);
            }
        }
        EngFncs->destroy_list(objects);
    }

    /* Avoid minor collision with containers */
    FOREACH_STORE(container, container_infos) {
        if (container->md_minor >= 0 && container->md_minor < minor_table_len) {
            hold_table[container->md_minor] = true;
            LOG_DETAILS("md minor %i is occupied by container %s\n", container->md_minor, container->evms_data->name);
        }
    }

    const u_int8_t minor_start = 127;
    u_int8_t minor = minor_start;
    if (!hold_table[minor])
        return LOG_EXIT((int)minor);

    --minor;
    for (; minor != minor_start; --minor)
        if (!hold_table[minor])
            return LOG_EXIT((int)minor);

    LOG_ERROR("Error getting free md minor number\n");
    return LOG_EXIT(-1);
}

/* Find container with the same superblock */
static container_info_t* find_container(imsm_super_t* sb) {
    LOG_ENTRY();
    FOREACH_STORE(container, container_infos)
        if (sb_equal(container->sb, sb))
            return LOG_EXIT((void*)container);
    return LOG_EXIT((void*)NULL);
}

/* Find disk with the same superblock */
static disk_info_t* find_disk(imsm_super_t* sb) {
    LOG_ENTRY();

    container_info_t* container = find_container(sb);
    if (!container)
        return LOG_EXIT((void*)NULL);

    FOREACH_LIST(current, container->disks) {
        if (sb_equal(current->sb, sb)) {
            LOG_DETAILS("found neighbour disk[%i]:%-16s\n", current->index, current->disk->serial);
            return LOG_EXIT((void*)current);
        }
    }
    return LOG_EXIT((void*)NULL);
}

/**
 * @brief Push new disk
 *
 * @param sb - superblock
 * @param disk - disk to push
 * @param evms_object - related evms storage object
 *
 * @return associated disk_info
 */
static disk_info_t* push_disk(imsm_super_t* sb, imsm_disk_t* disk, storage_object_t* evms_object) {
    LOG_ENTRY();

    disk_info_t* disk_info = disk_infos.data + disk_infos.count;
    disk_info->sb = sb;
    disk_info->disk = disk;
    disk_info->evms_data = evms_object;
    disk_info->index = (u_int8_t)(disk - sb->disk);

    /* Attach as neighbour */
    disk_info_t* neighbour = find_disk(disk_info->sb);
    if (neighbour) {
        /* Keep index order */
        disk_info_t* prev = NULL;
        while (neighbour && neighbour->index < disk_info->index) {
            LOG_DETAILS("skip neighbour disk[%i]:%-16s\n", neighbour->index, neighbour->disk->serial);
            prev = neighbour;
            neighbour = neighbour->next_tied;
        }
        if (prev)
            prev->next_tied = disk_info;

        LOG_DETAILS("pasting disk[%i]:%-16s before disk[%i]:%-16s\n",
                    disk_info->index,
                    disk_info->disk->serial,
                    neighbour ? neighbour->index : -1,
                    neighbour ? (char*)neighbour->disk->serial : "null");

        disk_info->next_tied = neighbour;
    }

    //evms_object->plugin = my_plugin_record;
    evms_object->consuming_private_data = disk_info;

    ++disk_infos.count;
    return LOG_EXIT((void*)disk_info);
}

static container_info_t* push_container(imsm_super_t* sb) {
    LOG_ENTRY();

    container_info_t* container = container_infos.data + container_infos.count;
    container->sb = sb;
    container->disks = NULL;
    container->volumes = NULL;
    container->md_minor = -1; /* NOT ASSINGED */
    container->active = false;

    char name[32];
    snprintf(name, sizeof(name) - 1, CONTAINER_PREFIX"%zu", container_infos.count);
    if (EngFncs->allocate_container(name, &container->evms_data)) {
        LOG_ERROR("Cannot allocate memory for storage container\n");
        return LOG_EXIT((void*)NULL);
    }
    storage_container_t* evms_container = container->evms_data;
    evms_container->plugin = my_plugin_record;
    evms_container->private_data = container;

    ++container_infos.count;

    return LOG_EXIT((void*)container);
}

static container_info_t* get_container(imsm_super_t* sb) {
    LOG_ENTRY();
    container_info_t* container = find_container(sb);
    if (!container)
        container = push_container(sb);
    return LOG_EXIT((void*)container);
}

static volume_info_t* find_volume(container_info_t* container, imsm_dev_t* dev) {
    LOG_ENTRY();
    FOREACH_LIST(vol_info, container->volumes)
        if (strncmp((const char*)vol_info->dev->volume, (const char*)dev->volume, sizeof(vol_info->dev->volume)) == 0)
            return LOG_EXIT((void*)vol_info);
    return LOG_EXIT((void*)NULL);
}

static char* find_usable_mddev_name(volume_info_t* vol, storage_container_t* evms_container);

static volume_info_t* push_volume(container_info_t* container, imsm_dev_t* dev) {
    LOG_ENTRY();
    volume_info_t* vol_info = volume_infos.data + volume_infos.count;
    vol_info->dev = dev;
    vol_info->container = container;
    vol_info->next_tied = NULL;
    vol_info->evms_data = NULL;
    vol_info->index = dev->my_vol_raid_dev_num - 1;

    volume_info_t* neighbour = container->volumes;
    if (neighbour) {
        /* Keep index order */
        volume_info_t* prev = NULL;
        while (neighbour && neighbour->index < vol_info->index) {
            prev = neighbour;
            neighbour = neighbour->next_tied;
        }
        if (prev)
            prev->next_tied = vol_info;
        vol_info->next_tied = neighbour;
    }

    /* Set list head if this disk have a smaller index */
    if (!container->volumes || container->volumes->index > vol_info->index)
        container->volumes = vol_info;

    /* Create EVMS data */
    char name[256];
    autocleanup(char*) usable_dev_name = find_usable_mddev_name(vol_info, container->evms_data);

    if (usable_dev_name)
        strncpy(name, usable_dev_name, sizeof(name) - 1);
    else
        make_evms_volume_name(name, sizeof(name) - 1, vol_info);

    if (EngFncs->allocate_region(name, &vol_info->evms_data)) {
        LOG_ERROR("Cannot allocate memory for storage container\n");
        return LOG_EXIT((void*)NULL);
    }
    storage_object_t* evms_vol = vol_info->evms_data;
    evms_vol->plugin = my_plugin_record;
    evms_vol->private_data = vol_info;
    evms_vol->data_type = DATA_TYPE;
    evms_vol->producing_container = container->evms_data;
    evms_vol->size = get_imsm_dev_size(dev);
    evms_vol->dev_major = MD_MAJOR;
    evms_vol->dev_minor = -1; /* NOT ASSIGNED */

    /* Set dummy geometry for segment managers (GPT/MBR) */
    evms_vol->geometry.bytes_per_sector = 512;
    evms_vol->geometry.heads = 255;
    evms_vol->geometry.sectors_per_track = 63;

    if (usable_dev_name) {
        evms_vol->dev_minor = get_mddev_minor(usable_dev_name);
        evms_vol->flags |= SOFLAG_ACTIVE;
        evms_vol->flags &= ~(u_int32_t)SOFLAG_NEEDS_ACTIVATE;
    }

    ++volume_infos.count;

    return LOG_EXIT((void*)vol_info);
}

typedef enum info_type { info_disk = 0, info_volume } info_type_e;

static u_int8_t get_private_data_info_index(storage_object_t* object, info_type_e type) {
    return type == info_disk ? ((disk_info_t*)object->consuming_private_data)->index
                             : ((volume_info_t*)object->private_data)->index;
}

static list_anchor_t get_container_bindings(storage_container_t* evms_container, info_type_e type) {
    return type == info_disk ? evms_container->objects_consumed : evms_container->objects_produced;
}

static void set_related_container(storage_object_t* obj, storage_container_t* container, info_type_e type) {
    if (type == info_disk)
        obj->consuming_container = container;
    else
        obj->producing_container = container;
}

static void
attach_some_to_container(storage_object_t* evms_object, storage_container_t* evms_container, info_type_e type) {
    LOG_ENTRY();
    LOG_DEBUG("attaching %s %s to container %s\n",
              type == info_disk ? "disk" : "volume",
              evms_object->name,
              evms_container->name);

    uint8_t index = get_private_data_info_index(evms_object, type);
    list_anchor_t container_bindings = get_container_bindings(evms_container, type);
    insert_flags_t flag = INSERT_AFTER;
    storage_object_t* iter_evms_object;
    list_element_t iter;

    LIST_FOR_EACH(container_bindings, iter, iter_evms_object) {
        if (index < get_private_data_info_index(iter_evms_object, type)) {
            flag = INSERT_BEFORE;
            break;
        }
    }

    set_related_container(evms_object, evms_container, type);
    EngFncs->insert_thing(container_bindings, evms_object, flag, iter);

    LOG_EXIT_VOID();
}

/**
 * @brief Creates internal structures
 *
 * @param sb - allocated superblock (lifetime managed by this function)
 * @param object - storage object
 *
 * @return 0 on success, any other code on error
 */
static int discover_disk(imsm_super_t* sb, storage_object_t* object, list_anchor_t output_objects) {
    LOG_ENTRY();

    /* Check capabilities */
    const char* no_platform_env = getenv("IMSM_NO_PLATFORM");
    boolean check_platform = !no_platform_env || strcmp(no_platform_env, "1") != 0;

    if (check_platform) {
        controller_info_t* controller = get_controller(object);
        if (!controller) {
            LOG_WARNING("Cannot find controller for device %s\n", object->dev_name);
            return LOG_EXIT(EINVAL);
        }

        for (u_int8_t i = 0; i < sb->num_raid_devs; ++i) {
            imsm_dev_t* dev = get_imsm_dev(sb, i);
            imsm_map_t* map = get_imsm_map(dev, MAP_0);

            if (!check_level_capability(controller->capabilities, map->raid_level)) {
                LOG_WARNING("Device %s wants raid%i, but controller %s has no capability for that\n",
                            object->dev_name,
                            (int)map->raid_level,
                            controller->sysfs_path);
                return LOG_EXIT(EINVAL);
            }
        }
    }

    /* Try find same sb and replace current by found */
    FOREACH_STORE(container, container_infos) {
        imsm_super_t* found = container->sb;
        if (sb_equal(found, sb)) {
            EngFncs->engine_free(sb);
            sb = found;
        }
    }

    /* Read the serial to identify the disk */
    int fd = EngFncs->open_object(object, O_RDONLY);
    u_int8_t serial[MAX_RAID_SERIAL_LEN];
    if (imsm_read_serial(fd, object->dev_name, serial, MAX_RAID_SERIAL_LEN))
        return LOG_EXIT(EINVAL);
    EngFncs->close_object(object, fd);

    imsm_disk_t* current_disk = NULL;
    for (u_int8_t i = 0; i < sb->num_disks; ++i) {
        imsm_disk_t* disk = sb->disk + i;
        if (memcmp(disk->serial, serial, sizeof(serial)) == 0) {
            current_disk = disk;

            char str[MAX_RAID_SERIAL_LEN + 1];
            strncpy(str, disk->serial, MAX_RAID_SERIAL_LEN);
            str[MAX_RAID_SERIAL_LEN] = '\0';
            LOG_DETAILS("Found disk with serial: %-16s index: %i\n", str, i);
        }
    }

    if (!current_disk) {
        LOG_WARNING("Cannot find the disk %-16s in imsm superblock\n", serial);
        return LOG_EXIT(EINVAL);
    }

    /* Disk setup */
    disk_info_t* disk_info = push_disk(sb, current_disk, object);
    container_info_t* container = get_container(sb);
    attach_some_to_container(disk_info->evms_data, container->evms_data, info_disk);

    /* Set list head if this disk have a smaller index */
    if (!container->disks || container->disks->index > disk_info->index) {
        char str[MAX_RAID_SERIAL_LEN + 1];
        strncpy(str, disk_info->disk->serial, MAX_RAID_SERIAL_LEN);
        str[MAX_RAID_SERIAL_LEN] = '\0';

        LOG_DETAILS("set new head: disk[%i]:%-16s\n", disk_info->index, str);
        container->disks = disk_info;
    }

    for (u_int8_t i = 0; i < sb->num_raid_devs; ++i) {
        volume_info_t* vol = find_volume(container, get_imsm_dev(sb, i));
        if (!vol) {
            vol = push_volume(container, get_imsm_dev(sb, i));
            EngFncs->insert_thing(output_objects, vol->evms_data, INSERT_AFTER, NULL);
            attach_some_to_container(vol->evms_data, container->evms_data, info_volume);
        }
    }

    return LOG_EXIT(0);
}

static int imsm_setup_evms_plugin(engine_functions_t* functions) {
    EngFncs = functions;
    LOG_ENTRY();
    return LOG_EXIT(0);
}

static void imsm_cleanup_evms_plugin(void) {
    LOG_ENTRY();

    FOREACH_STORE(disk, disk_infos)
        disk->evms_data->consuming_private_data = NULL;

    FOREACH_STORE(container, container_infos) {
        FOREACH_LIST(vol, container->volumes)
            EngFncs->free_region(vol->evms_data);

        EngFncs->free_container(container->evms_data);
        EngFncs->engine_free(container->sb);

        memset(&container_infos, 0, sizeof(container_infos));
        memset(&disk_infos, 0, sizeof(container_infos));
        memset(&volume_infos, 0, sizeof(container_infos));
    }

    LOG_EXIT_VOID();
}

sector_count_t imsm_mpb_sectors(imsm_super_t* sb, u_int64_t size) {
    LOG_ENTRY();

    LOG_DEBUG("MPB size: %u, rounding up to %ld\n", sb->mpb_size, size);
    size--;
    sector_count_t sector_count = ((sb->mpb_size + size) & ~size) / size; // round up

    return LOG_EXIT(sector_count);
}

static int imsm_calc_sb_csum(imsm_super_t* sb) {
    LOG_ENTRY();

    u_int32_t end = sb->mpb_size / sizeof(end);
    u_int32_t* p = (u_int32_t*)sb;
    u_int32_t sum = 0;

    LOG_DEBUG("%d\n", sb->mpb_size);
    while (end--) {
        sum += DISK_TO_CPU32(*p);
        LOG_DEBUG("%d: sum += %#x = %#x\n", end, *p, sum);
        p++;
    }

    sum = sum - DISK_TO_CPU32(sb->check_sum);
    return LOG_EXIT(sum);
}

static int imsm_validate_csum(imsm_super_t* sb) {
    LOG_ENTRY();

    int expected = DISK_TO_CPU32(sb->check_sum);
    int actual = imsm_calc_sb_csum(sb);

    if (actual != expected) {
        LOG_WARNING("Bad Checksum: got %#x, expected %#x\n", actual, expected);
        return LOG_EXIT(ENXIO);
    }
    return LOG_EXIT(0);
}

static int imsm_validate_sb(imsm_super_t* sb) {
    LOG_ENTRY();

    if (strncmp((char*)sb->sig, MPB_SIGNATURE, MPB_SIG_LEN) != 0) {
        LOG_EXTRA("Bad signature: got %s, expected %s\n", (char*)sb->sig, MPB_SIGNATURE);
        return LOG_EXIT(ENXIO);
    }

    return LOG_EXIT(0);
}

static int imsm_read_data(storage_object_t* obj, u_int64_t loc, void* buf, u_int32_t sects) {
    LOG_ENTRY();

    if (obj->data_type != DATA_TYPE) {
        LOG_DETAILS("Object not data type, skipping %s.\n", obj->name);
        return LOG_EXIT(0);
    }

    LOG_DEBUG("Looking for IMSM data at %" PRIu64 " on %s.\n", loc, obj->name);

    int rc = 0;
    if (obj->flags & SOFLAG_ACTIVE) {
        /* Make sure that we read what's on disk */
        int fd = EngFncs->open_object(obj, O_RDONLY | O_SYNC);
        if (fd <= 0) {
            LOG_ERROR("Error opening object %s.\n", obj->name);
            return LOG_EXIT(EIO);
        }

        int read_bytes = EngFncs->read_object(obj, fd, buf, sects << EVMS_VSECTOR_SIZE_SHIFT, loc << EVMS_VSECTOR_SIZE_SHIFT);
        if (read_bytes != (int32_t)(sects << EVMS_VSECTOR_SIZE_SHIFT)) {
            LOG_ERROR("READ from %s failed, rc=%d.\n", obj->name, read_bytes);
            rc = EIO;
        }

        EngFncs->close_object(obj, fd);
    }
    else if (READ(obj, loc, sects, (char*)buf))
        rc = EIO;

    return LOG_EXIT(rc);
}

static void imsm_print_disk(imsm_disk_t* d) {
    LOG_ENTRY();

    char str[MAX_RAID_SERIAL_LEN + 1];
    strncpy(str, (char*)d->serial, MAX_RAID_SERIAL_LEN);
    str[MAX_RAID_SERIAL_LEN] = '\0';

    LOG_DEBUG("Sizeof structure: %zu bytes.\n", sizeof(*d));
    LOG_DEBUG("serial:          %s\n", str);
    LOG_DEBUG("total_blocks_lo: %u\n", d->total_blocks_lo);
    LOG_DEBUG("scsi_id:         %u\n", d->scsi_id);
    LOG_DEBUG("status:          %u\n", d->status);
    LOG_DEBUG("owner_cfg_num:   %u\n", d->owner_cfg_num);
    LOG_DEBUG("total_blocks_hi: %u\n", d->total_blocks_hi);

    LOG_EXIT_VOID();
}

static void imsm_print_sb(void* super) {
    imsm_super_t* sb = (imsm_super_t*)super;
    int i;
    LOG_ENTRY();

    LOG_DEBUG("Sizeof structure: %zu bytes.\n", sizeof(*sb));
    LOG_DEBUG("sig:                   %s\n", sb->sig);
    LOG_DEBUG("check_sum:             %u\n", sb->check_sum);
    LOG_DEBUG("mpb_size:              %u\n", sb->mpb_size);
    LOG_DEBUG("family_num:            %u\n", sb->family_num);
    LOG_DEBUG("generation_num:        %u\n", sb->generation_num);
    LOG_DEBUG("error_log_size:        %u\n", sb->error_log_size);
    LOG_DEBUG("attributes:            %u\n", sb->attributes);
    LOG_DEBUG("num_disks:             %u\n", sb->num_disks);
    LOG_DEBUG("num_raid_devs:         %u\n", sb->num_raid_devs);
    LOG_DEBUG("error_log_pos:         %u\n", sb->error_log_pos);
    LOG_DEBUG("cache_size:            %u\n", sb->cache_size);
    LOG_DEBUG("orig_family_num:       %u\n", sb->orig_family_num);
    LOG_DEBUG("pwr_cycle_count:       %u\n", sb->pwr_cycle_count);
    LOG_DEBUG("bbm_log_size:          %u\n", sb->bbm_log_size);
    LOG_DEBUG("num_raid_devs_created: %u\n", sb->num_raid_devs_created);
    LOG_DEBUG("creation_time:         %" PRIu64 "\n", sb->creation_time);
    for (i = 0; i < sb->num_disks; i++) {
        imsm_print_disk(&sb->disk[i]);
    }

    LOG_EXIT_VOID();
}

static int imsm_read_sb(storage_object_t* obj, imsm_super_t** super) {
    LOG_ENTRY();

    autocleanup(imsm_super_t*) sb = EngFncs->engine_alloc(MD_IMSM_BYTES);
    if (!sb)
        return LOG_EXIT(ENOMEM);

    int rc = imsm_read_data(obj, obj->size - 2, sb, MD_IMSM_SECTORS);
    if (rc)
        return LOG_EXIT(rc);

    if ((rc = imsm_validate_sb(sb)))
        return LOG_EXIT(rc);

    sector_count_t sectors = imsm_mpb_sectors(sb, MD_IMSM_SECTOR_SIZE) - 1;
    if (sectors && (rc = imsm_read_data(obj, obj->size - (2 + sectors), ((void*)sb) + MD_IMSM_SECTOR_SIZE, sectors)))
        return LOG_EXIT(rc);

    if ((rc = imsm_validate_csum(sb)))
        return LOG_EXIT(rc);

    imsm_print_sb(sb);
    *super = sb;
    sb = NULL;

    return LOG_EXIT(rc);
}

static int discover_imsm_disks(list_anchor_t input_objects, list_anchor_t output_objects) {
    LOG_ENTRY();

    int count = 0;

    storage_object_t* object;
    list_element_t itr1, itr2;
    LIST_FOR_EACH_SAFE(input_objects, itr1, itr2, object) {
        if (object->data_type != DATA_TYPE) {
            LOG_DEBUG("Skipping object %s - not DATA_TYPE.\n", object->name);
            continue;
        }

        imsm_super_t* sb = NULL;
        if (!imsm_read_sb(object, &sb)) {
            if (!discover_disk(sb, object, output_objects)) {
                EngFncs->remove_element(itr1);
                ++count;
            }

            LOG_DEBUG("Containers:\n");
            FOREACH_STORE(container, container_infos) {
                LOG_DEBUG("    Disks:\n");
                FOREACH_LIST(disk, container->disks) {
                    char str[MAX_RAID_SERIAL_LEN + 1];
                    strncpy(str, disk->disk->serial, MAX_RAID_SERIAL_LEN);
                    str[MAX_RAID_SERIAL_LEN] = '\0';

                    LOG_DEBUG("        disk[%i]: %-16s\n", disk->index, str);
                }

                LOG_DEBUG("    Volumes:\n");
                FOREACH_LIST(vol, container->volumes)
                    LOG_DEBUG("        volume[%i]: %-16s\n", vol->index, vol->dev->volume);
            }
        }
    }

    return LOG_EXIT(count);
}

static int imsm_discover(list_anchor_t input_objects, list_anchor_t output_objects, boolean final_call) {
    LOG_ENTRY();
    LOG_DEBUG("Beginning IMSM Discovery\n");

    int count = discover_imsm_disks(input_objects, output_objects);

    /* Copy the remaining objects to the output list. */
    EngFncs->merge_lists(output_objects, input_objects, NULL, NULL);

    LOG_DEBUG("IMSM Discovery complete.\n");
    return LOG_EXIT(count);
}

static int imsm_discard(list_anchor_t regions) {
    storage_object_t* region;
    list_element_t iter1, iter2;

    LOG_ENTRY();

    LIST_FOR_EACH_SAFE(regions, iter1, iter2, region) {
        if (region->data_type != DATA_TYPE) {
            LOG_DEBUG("Cannot discard freespace region %s.\n", region->name);
            continue;
        }

        LOG_DEBUG("Discarding region %s.\n", region->name);
        /* TODO: do discard things */
    }

    return LOG_EXIT(0);
}

static int imsm_delete(storage_object_t *obj, list_anchor_t UNUSED(child_objects)) {
    int rc = 0;
    LOG_ENTRY();

    /* pass */

    LOG_EXIT_INT(rc);
    return rc;
}

/**
 * imsm_commit_changes
 *
 * All IMSM data commits are done in imsm_commit_container_changes.
 **/
static int imsm_commit_changes(storage_object_t* region, commit_phase_t UNUSED(phase)) {
    LOG_ENTRY();
    /* Nothing to commit */
    region->flags &= ~(SOFLAG_DIRTY | SOFLAG_NEW);
    return LOG_EXIT(0);
}

typedef enum {
    imsm_io_read = 0,
    imsm_io_write
} imsm_io_e;

static int imsm_dev_io(storage_object_t* region, lsn_t lsn, sector_count_t count, void* buffer, imsm_io_e op) {
    LOG_ENTRY();

    if (!count)
        return LOG_EXIT(0);

    autocleanup(char*) vol_dev_name = strdup_cat("/dev/", region->name);
    int fd = open(vol_dev_name, O_SYNC | O_RDWR);
    if (fd <= 0) {
        LOG_ERROR("Error opening imsm object %s.\n", region->name);
        return LOG_EXIT(ENODEV);
    }

    size_t bytes = count << EVMS_VSECTOR_SIZE_SHIFT;
    size_t offset = lsn << EVMS_VSECTOR_SIZE_SHIFT;

    int rc = 0;
    off64_t curr_offset = lseek64(fd, offset, SEEK_SET);
    if ((size_t)curr_offset != offset) {
        LOG_WARNING("lseek64 to offset %" PRIu64 " on object %s failed with error code %d: %s\n",
                    offset,
                    region->name,
                    errno,
                    strerror(errno));
        rc = EIO;
        goto out;
    }

    void *old_buffer = buffer;

    if (buffer == NULL) {
        buffer = EngFncs->engine_alloc(bytes);
    }

    if (buffer == NULL) {
        LOG_ERROR("Error allocating buffer\n");
        rc = ENOMEM;
        goto out;
    }

    ssize_t cur_bytes = op == imsm_io_read ? read(fd, buffer, bytes) : write(fd, buffer, bytes);
    if (cur_bytes < 0) {
        LOG_WARNING("Read from object %s failed with error code %d: %s\n", region->name, errno, strerror(errno));
        rc = EIO;
        goto out;
    }

out:
    if (old_buffer == NULL) {
        EngFncs->engine_free(buffer);
        buffer = NULL;
    }

    close(fd);
    return LOG_EXIT(rc);
}

static int imsm_activate(storage_object_t* evms_vol);

/**
 * imsm_read
 *
 * Process a read request on this region at the given starting sector for the
 * given number of sectors.
 **/
static int imsm_read(storage_object_t* region, lsn_t lsn, sector_count_t count, void* buffer) {
    LOG_ENTRY();
    LOG_DEBUG("Read request on region %s, sector %" PRIu64 ", count %" PRIu64 ".\n", region->name, lsn, count);

    if (!count)
        return LOG_EXIT(0);

    if ((region->flags & SOFLAG_ACTIVE) == 0)
        imsm_activate(region);

    int rc = imsm_dev_io(region, lsn, count, buffer, imsm_io_read);
    return LOG_EXIT(rc);
}

/**
 * imsm_write
 *
 * Process a write request on this region at the given starting sector for the
 * given number of sectors.
 **/
static int imsm_write(storage_object_t* region, lsn_t lsn, sector_count_t count, void* buffer) {
    LOG_ENTRY();
    LOG_DEBUG("Write request on region %s, sector %" PRIu64 ", count %" PRIu64 ".\n", region->name, lsn, count);

    if (!count)
        return LOG_EXIT(0);

    if ((region->flags & SOFLAG_ACTIVE) == 0)
        imsm_activate(region);

    int rc = imsm_dev_io(region, lsn, count, buffer, imsm_io_write);
    return LOG_EXIT(rc);
}

/**
 * imsm_add_sectors_to_kill_list
 *
 * Process a kill-sectors request on this region at the given starting sector
 * for the given number of sectors.
 **/
static int imsm_add_sectors_to_kill_list(storage_object_t* region, lsn_t lsn, sector_count_t count) {
    LOG_ENTRY();
    LOG_DEBUG("Kill-sectors on region %s, sector %" PRIu64 ", count %" PRIu64 ".\n", region->name, lsn, count);

    if (!count)
        return LOG_EXIT(0);

    if ((region->flags & SOFLAG_ACTIVE) == 0)
        imsm_activate(region);

    int rc = imsm_dev_io(region, lsn, count, NULL, imsm_io_write);
    return LOG_EXIT(rc);
}

static int imsm_get_info(storage_object_t* region, char* info_name, extended_info_array_t** info) {
    LOG_ENTRY();

    if (info_name) {
        LOG_ERROR("no support for extra region information about \"%s\"\n", info_name);
        return LOG_EXIT(EINVAL);
    }

    int rc = get_volume_info(region, info);
    return LOG_EXIT(rc);
}

static int imsm_get_plugin_info(char* info_name, extended_info_array_t** info_array) {
    LOG_ENTRY();

    if (info_name) {
        LOG_ERROR("no support for extra plugin information about \"%s\"\n", info_name);
        return LOG_EXIT(EINVAL);
    }

    int rc = get_plugin_info(info_array);
    return LOG_EXIT(rc);
}

static int imsm_can_activate(storage_object_t* evms_vol) {
    /* TODO: implement this */
    LOG_ENTRY();
    return LOG_EXIT(0);
}

static int assemble_container(storage_container_t* evms_container) {
    LOG_ENTRY();

    int dev_minor = find_free_md_minor();
    autocleanup(char*) container_link_path = strdup_cat("/dev/imsm/", evms_container->name);
    autocleanup(char*) md_dev_path = make_mddev_path(dev_minor);

    list_anchor_t evms_disks = evms_container->objects_consumed;
    size_t disks_count = EngFncs->list_count(evms_disks);
    char* args[4 + disks_count + 1];

    args[0] = "/sbin/mdadm";
    args[1] = "--assemble";
    args[2] = "--metadata=imsm";
    args[3] = md_dev_path;
    args[4 + disks_count] = NULL;
    size_t arg_index = 4;

    storage_object_t* evms_disk;
    list_element_t iter;
    LIST_FOR_EACH(evms_disks, iter, evms_disk)
        args[arg_index++] = make_dev_name(evms_disk->name);

    //setenv("IMSM_NO_PLATFORM", "1", 0);
    int rc = fork_and_execvp(args, NULL);
    if (rc != 0)
        goto out;

    rc = ensure_md_symlink(container_link_path, dev_minor);
    if (rc == 0) {
        container_info_t* container = evms_container->private_data;
        container->active = true;
        container->md_minor = dev_minor;
    }

out:
    for (size_t i = 4; i < 4 + disks_count; ++i)
        EngFncs->engine_free(args[i]);

    return LOG_EXIT(rc);
}

static boolean is_container_assembled(storage_container_t* evms_container) {
    container_info_t* container = evms_container->private_data;
    return container->active;
}

static int assemble_volume(storage_object_t* evms_vol) {
    LOG_ENTRY();

    storage_container_t* evms_container = evms_vol->producing_container;
    int dev_minor = find_free_md_minor();
    autocleanup(char*) vol_dev_link_path = strdup_cat("/dev/", evms_vol->name);
    autocleanup(char*) container_dev_link_path = strdup_cat("/dev/imsm/", evms_container->name);
    autocleanup(char*) md_dev_path = make_mddev_path(dev_minor);

    char* args[5] = {"/sbin/mdadm", "--assemble", md_dev_path, container_dev_link_path, NULL};
    //setenv("IMSM_NO_PLATFORM", "1", 0);
    int rc = fork_and_execvp(args, NULL);
    if (rc != 0)
        goto out;

    rc = ensure_md_symlink(vol_dev_link_path, dev_minor);
    if (rc == 0) {
        evms_vol->dev_minor = dev_minor;
        strncpy(evms_vol->dev_name, vol_dev_link_path, sizeof(evms_vol->dev_name) - 1);
    }

out:
    return LOG_EXIT(rc);
}

static int reverse_volumes_traverse(volume_info_t* list, int(*handler)(storage_object_t*)) {
    LOG_ENTRY();

    if (!list)
        return LOG_EXIT(0);

    int rc = reverse_volumes_traverse(list->next_tied, handler);
    if (rc == 0)
        rc = handler(list->evms_data);

    return LOG_EXIT(rc);
}

static int imsm_activate(storage_object_t* evms_vol) {
    LOG_ENTRY();

    if (evms_vol->flags & SOFLAG_ACTIVE) {
        /* TODO: handle already active */
        return LOG_EXIT(0);
    }

    storage_container_t* evms_container = evms_vol->producing_container;
    if (!is_container_assembled(evms_container)) {
        int rc = assemble_container(evms_container);
        if (rc != 0)
            return LOG_EXIT(rc);
    }

    /* Assemble all volumes in container 
     *
     * XXX: assembling specified volume only is better, but there is no way to do that with mdadm :(
     * This is not working because mdadm assembles volumes one by one started from the last volume in superblock.
     * For the instance:
     *
     *     $ mdadm --assemble Volume0 /dev/md/imsm0          <- trying to assemble Volume0, the first member
     *     $ mdadm --detail /dev/md/Volume0 | grep member
     *               Container : /dev/md/imsm0, member 2     <- wtf, did we just assembled the third member!?
     *
     * (but maybe there is a way...)
     */
    container_info_t* container = evms_container->private_data;
    int rc = reverse_volumes_traverse(container->volumes, assemble_volume);

    if (rc == 0) {
        storage_object_t* evms_vol;
        list_element_t iter;
        LIST_FOR_EACH(evms_container->objects_produced, iter, evms_vol) {
            evms_vol->flags |= SOFLAG_ACTIVE;
            evms_vol->flags &= ~((u_int32_t)SOFLAG_NEEDS_ACTIVATE);
        }
    }

    return LOG_EXIT(rc);
}

static int imsm_can_deactivate(storage_object_t* evms_vol) {
    /* TODO: implement this */
    LOG_ENTRY();
    return LOG_EXIT(0);
}

static int stop_volume(storage_object_t* evms_vol) {
    LOG_ENTRY();

    if (evms_vol->dev_minor == -1 || !(evms_vol->flags & SOFLAG_ACTIVE))
        return LOG_EXIT(0);

    char* args[] = {"/sbin/mdadm", "--stop", evms_vol->dev_name, NULL};
    int rc = fork_and_execvp(args, NULL);
    if (rc == 0) {
        evms_vol->dev_minor = -1;
        evms_vol->dev_name[0] = '\0';
    }

    return LOG_EXIT(rc);
}

static int stop_container(storage_container_t* evms_container) {
    LOG_ENTRY();

    if (!((container_info_t*)evms_container->private_data)->active)
        return LOG_EXIT(0);

    autocleanup(char*) container_link_path = strdup_cat("/dev/imsm/", evms_container->name);

    char* args[] = {"/sbin/mdadm", "--stop", container_link_path, NULL};
    int rc = fork_and_execvp(args, NULL);
    if (rc == 0) {
        container_info_t* container = evms_container->private_data;
        container->active = false;
        container->md_minor = -1;
    }

    return LOG_EXIT(rc);
}


static int imsm_deactivate(storage_object_t* evms_vol) {
    LOG_ENTRY();

    if ((evms_vol->flags & SOFLAG_ACTIVE) == 0) {
        return LOG_EXIT(0);
    }

    storage_container_t* evms_container = evms_vol->producing_container;
    if (!evms_container) {
        return LOG_EXIT(0);
    }

    container_info_t* container = evms_container->private_data;
    if (!container || !container->active) {
        return LOG_EXIT(0);
    }

    int rc = reverse_volumes_traverse(container->volumes, stop_volume);
    if (rc == 0) {
        storage_object_t* evms_vol;
        list_element_t iter;
        LIST_FOR_EACH(evms_container->objects_produced, iter, evms_vol)
            evms_vol->flags &= ~(u_int32_t)SOFLAG_ACTIVE;

        rc = stop_container(evms_container);
    }

    return LOG_EXIT(rc);
}

#define NOT_IMPLEMENTED NULL

static plugin_functions_t imsm_functions = {
    .setup_evms_plugin          = imsm_setup_evms_plugin,
    .cleanup_evms_plugin        = imsm_cleanup_evms_plugin,
    .can_delete                 = NOT_IMPLEMENTED,
    .can_expand                 = NOT_IMPLEMENTED, // imsm_can_expand,
    .can_shrink                 = NOT_IMPLEMENTED, // imsm_can_shrink,
    .discover                   = imsm_discover,
    .create                     = NOT_IMPLEMENTED, // imsm_create,
    .discard                    = imsm_discard,
    .delete                     = imsm_delete,
    .expand                     = NOT_IMPLEMENTED, // imsm_expand,
    .shrink                     = NOT_IMPLEMENTED, // imsm_shrink,
    .commit_changes             = imsm_commit_changes,
    .can_activate               = imsm_can_activate,
    .activate                   = imsm_activate,
    .can_deactivate             = imsm_can_deactivate,
    .deactivate                 = imsm_deactivate,
    .get_option_count           = NOT_IMPLEMENTED, // imsm_get_option_count,
    .init_task                  = NOT_IMPLEMENTED, // imsm_init_task,
    .set_objects                = NOT_IMPLEMENTED, // imsm_set_objects,
    .set_option                 = NOT_IMPLEMENTED, // imsm_set_option,
    .get_info                   = imsm_get_info,
    .set_info                   = NOT_IMPLEMENTED, // imsm_set_info,
    .get_plugin_info            = imsm_get_plugin_info,
    .read                       = imsm_read,
    .write                      = imsm_write,
    .add_sectors_to_kill_list   = imsm_add_sectors_to_kill_list,
    .get_plugin_functions       = NOT_IMPLEMENTED, // imsm_get_functions,
    .plugin_function            = NOT_IMPLEMENTED, // imsm_function,
    .backup_metadata            = NOT_IMPLEMENTED, // imsm_backup_metadata,
};

static int imsm_can_delete_container(storage_container_t* container) {
    /* XXX: not implemented - cannot be deleted */
    LOG_ENTRY();
    return LOG_EXIT(EBUSY);
}

static int imsm_create_container(list_anchor_t objects, option_array_t* options, storage_container_t** new_container) {
    /* XXX: not implemented - cannot be created */
    LOG_ENTRY();
    return LOG_EXIT(EINVAL);
}

static int imsm_discard_container(storage_container_t *container) {
    /* XXX: not implemented - cannot be discarded */
    LOG_ENTRY();
    return LOG_EXIT(EINVAL);
}

static int imsm_delete_container(storage_container_t* container, list_anchor_t objects_consumed) {
    /* XXX: not implemented - cannot be deleted */
    LOG_ENTRY();
    return LOG_EXIT(EINVAL);
}

static int imsm_commit_container_changes(storage_container_t* container, commit_phase_t phase) {
    /* XXX: not implemented - cannot be commited */
    LOG_ENTRY();
    return LOG_EXIT(EINVAL);
}

static int imsm_get_container_info(storage_container_t* container, char* name, extended_info_array_t** info) {
    /* XXX: not implemented - info cannot be provided */
    LOG_ENTRY();
    return LOG_EXIT(EINVAL);
}

static container_functions_t imsm_container_functions = {
    .can_delete_container = imsm_can_delete_container,
    .can_expand_container = NOT_IMPLEMENTED,      // imsm_can_expand_container,
    .can_shrink_container = NOT_IMPLEMENTED,      // imsm_can_shrink_container,
    .can_expand_container_by = NOT_IMPLEMENTED,   // imsm_can_expand_container_by,
    .can_shrink_container_by = NOT_IMPLEMENTED,   // imsm_can_shrink_container_by,
    .create_container = imsm_create_container,
    .expand_container = NOT_IMPLEMENTED,          // imsm_expand_container,
    .shrink_container = NOT_IMPLEMENTED,          // imsm_shrink_container,
    .discard_container = imsm_discard_container,
    .delete_container = imsm_delete_container,
    .commit_container_changes = imsm_commit_container_changes,
    .get_container_info = imsm_get_container_info,
    .set_container_info = NOT_IMPLEMENTED,        // imsm_set_container_info,
    .get_plugin_functions = NOT_IMPLEMENTED,      // imsm_get_container_functions,
    .plugin_function = NOT_IMPLEMENTED,           // imsm_container_function,
    .backup_container_metadata = NOT_IMPLEMENTED, // imsm_backup_container_metadata,
};

plugin_record_t imsm_plugin = {
    .id = EVMS_IMSM_PLUGIN_ID,
    .version = {.major = MAJOR_VERSION, .minor = MINOR_VERSION, .patchlevel = PATCH_LEVEL},
    .required_engine_api_version = {.major = 15, .minor = 0, .patchlevel = 0},
    .required_plugin_api_version = {.plugin = {.major = 13, .minor = 1, .patchlevel = 0}},
    .required_container_api_version = {.major = 10, .minor = 1, .patchlevel = 0},
    .short_name = EVMS_IMSM_PLUGIN_SHORT_NAME,
    .long_name = EVMS_IMSM_PLUGIN_LONG_NAME,
    .oem_name = EVMS_ALT_LINUX_OEM_NAME,
    .functions = {.plugin = &imsm_functions},
    .container_functions = &imsm_container_functions,
    .discover_before_segments = TRUE,
};

plugin_record_t* evms_plugin_records[] = {&imsm_plugin, NULL};
